/*
** An abstract array type for C++
**
** Copyright (c) 1999 D. Richard Hipp
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
** 
** You should have received a copy of the GNU General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA  02111-1307, USA.
**
** Author contact information:
**   drh@hwaci.org
**   http://www.hwaci.com/drh/
*/
#include <string.h>
#include <stdio.h>
#include "array.h"

static char const rcsid[] = "@(#) $Header: /home/softools/cvs/softools/code/common/cked/src/array.cpp,v 1.4 2003/05/13 00:43:42 drh Exp $";

#if INTERFACE
/*
** object: SsArray
**
** This class implements an associative array.  The data is a pointer
** to anything.  The key type is undefined in this class (but it is
** defined in subclasses.)
**
** This is an abstract class. To be useful, the protected key handling
** functions must be defined in a subclass.
*/
class SsArray {
private:
  int nElem;              // Number of elements in the array 
  SsArrayElem *pFirst;    // First element in the array
  int nHtBucket;          // Number of hash table buckets 
  struct SsArray_ht {     // One such structure for each hash table bucket 
    int nCollide;           // Number of entries with this hash
    SsArrayElem *pChain;    // First entry with this hash
  } *aHt;                 // The hash table

protected:
  /*
   * The following functions are used to handle keys.  Subclasses must
   * define them appropriately.
   */
  virtual int xHash(const void*) const = 0;
  virtual int xCompare(const void*, const void*) const = 0;
  virtual void *xCopy(void*) const = 0;
  virtual void *xCopyData(void*) const = 0;
  virtual void xDelete(void*) const = 0;
  virtual void xDeleteData(void*) const = 0;

public:
  SsArray(){
    xReset();
  };
  virtual ~SsArray(){}
  int count() const {
    return nElem;
  };
  SsArrayElem *first() const { return pFirst; };
  void reset(){ xClear(); };
  void *operator[](void *key){ return find(key); };
};

/*
** object: SsArrayElem
**
** Each element of an SsArray is an instance of this class.  Only
** an SsArray object can create or destory one of these, but other
** objects can access an SsArrayElem in order to get at the data
** within an SsArray.
**
** The usual idiom for looping over all elements of an SsArray
** is the following:
**
** <blockquote><pre>
** SsArray array;
** SsArrayElem *p;
** // ...
** for(p=array.first(); p; p=p->next()){
**    SomeObject *pObj = (SomeObject*)p->data();
**    // Do something with pObj...
** }
** </pre></blockquote>
**
** The SsArrayElem object exists principly to support loops such as
** the one above.
*/
class SsArrayElem {
private:
  SsArrayElem *pNext, *pPrev; // Next and previous elements in the array
  void *pData;                // Data associated with this element
  void *pKey;                 // Key associated with this element
  SsArrayElem(){};            // Only an Array can create an ArrayElem
  ~SsArrayElem(){};           // Only an Array can destroy an ArrayElem
  friend class SsArray;
  
public:
  SsArrayElem *next() const { return pNext; };
  SsArrayElem *prev() const { return pPrev; };
  int isLast() const { return pNext==0; };
  int isFirst() const { return pPrev==0; };
  void *data() const { return pData; };
  void *key() const { return pKey; };
};

/*
** object: SsArrayOfPointer
**
** This subclass of SsArray uses pointers as the key.  Objects of this
** type can store an arbitrary number of pointers each keyed a different
** pointer.
*/
class SsArrayOfPointer : public SsArray {
protected:
  int xHash(const void *key) const { 
    return (int)key ^ ((int)key >> 6); 
  };
  int xCompare(const void *key1, const void *key2) const {
    return (int)key1 - (int)key2;
  };
  void *xCopy(void *key) const { return (void*)(int)key; };
  void *xCopyData(void *data) const { return (void*)(int)data; };
  void xDelete(void *key) const { return; };
  void xDeleteData(void *data) const { return; };
public:
  SsArrayOfPointer(){};
  virtual ~SsArrayOfPointer(){ reset(); };
};

/*
** object: SsArrayOfString
**
** This subclass of SsArray uses a null-terminated character string as
** the key.  For each element that is inserted into the array, the
** array makes a copy of the key into private memory.  The key is
** automatically deleted when the element is removed from the array or
** when the array itself is deleted.
**
** For objects of this type, case is significant when comparing keys.
**
** See also SsArrayOfStaticString.
*/
class SsArrayOfString : public SsArray {
private:
  static int strhash(const char *z){
    int h = 0;
    while( *z ){ h = h ^ (h<<5) ^ *(z++); }
    return  h & 0x7fffffff;
  }
protected:
  int xHash(const void *key) const { 
    return strhash((char*)key);
  };
  int xCompare(const void *key1, const void *key2) const {
    return strcmp((const char*)key1,(const char*)key2);
  };
  void *xCopy(void *key) const {
    char *z = new char[strlen((char*)key)+1];
    strcpy(z,(char*)key);
    return (void*)z;
  };
  void *xCopyData(void *data) const { return (void*)(int)data; };
  void xDelete(void *key) const { delete [] (char*)key; }
  void xDeleteData(void *data) const { return; };
public:
  SsArrayOfString(){
    return;
  };
  virtual ~SsArrayOfString(){ reset(); };
};

/*
** object: SsStringArrayOfString
**
** This subclass of SsArrayOfString uses a null-terminated character string
** as both the key and the data.  The array makes a copy of both the key
** and the data into private memory.
*/
class SsStringArrayOfString : public SsArrayOfString {
protected:
  void *xCopyData(void *data) const {
    char *z = new char[strlen((char*)data)+1];
    strcpy(z,(char*)data);
    return (void*)z;
  }
  void xDeleteData(void *data) const {delete [] (char *)data; }
public:
  SsStringArrayOfString(){
    return;
  }
  virtual ~SsStringArrayOfString(){ reset(); }
}

/*
** object: SsArrayOfStaticString
**
** This subclass of SsArray uses a null-terminated character string as
** the key.  The array does not make its own copy of the key.  It just
** stores a pointer to the key.  Thus it assumes that the key string
** is permanent.
**
** See also SsArrayOfString.
*/
class SsArrayOfStaticString : public SsArrayOfString {
protected:
  void *xCopy(void *key) const { return key; };
  void xDelete(void *key) const {};
public:
  SsArrayOfStaticString(){
    return;
  };
  virtual ~SsArrayOfStaticString(){ reset(); };
};

/*
** An array element and a set element differ only in that
** the data element itself is ignored; the key is the data.
** This is important, because for SsSetOfString, a private copy
** of the key is made, but the data may be invalid.
**
** The caller must be aware of this difference, and should never
** use the data() method for a Set.
*/
typedef SsArrayElem SsSetElem;

/*
** object: SsSetOfPointer
**
** SsSetOfPointer implements an unordered set of pointers.  Pointers can
** be added or removed from the set or an iterator can be run across the
** entire set.
**
** SsSetOfPointer uses an SsArrayOfPointer internally to implement the set.
**
** The idiom for looping over all elements of a set is the same as for
** looping over all elements of an array.  (See the documentation on
** SsArrayElem for additional information.)  We have:
**
** <blockquote><pre>
** SsSetOfPointer set;
** SsSetElem *p;
** // ...
** for(p=array.first(); p; p=p->next()){
**    SomeObject *pObj = (SomeObject*)p->data();
**    // Do something with pObj...
** }
** </pre></blockquote>
**
** The SsSetElem type is really just an alias for SsArrayElem.
*/
class SsSetOfPointer {
private:
  SsArrayOfPointer array;
public:
  SsSetOfPointer(){
    return;
  };
  ~SsSetOfPointer(){};
  int count() const { return array.count(); };
  SsSetElem *first() const { return array.first(); };
  void reset(){ array.reset(); };
  void insert(void *data){ array.insert(data,data); };
  int find(void *data) const { return array.find(data)!=0; };
  void remove(const void *data){ array.remove(data); };
  int operator[](void *data) const { return find(data); };
  void xfer(SsSetOfPointer *pOther){ array.xfer(&pOther->array); };
};

/*
** object: SsSetOfString
**
** SsSetOfString implements an unordered set of null-terminated strings.
** When new strings are inserted into this set, a copy is made of the
** string in memory obtained from the C++ "new" operator.  Thus ephemeral
** strings may be added to the set.  The set automatically releases the
** memory when the element is removed from the set or when the set is
** destroyed.
**
** See also SsArrayOfString.  SsSetOfString uses an SsArrayOfString
** internally in order to implement the set abtraction.
**
** See also SsSetOfPointer.
*/
class SsSetOfString {
private:
  SsArrayOfString array;
public:
  SsSetOfString(){
    return;
  };
  ~SsSetOfString(){};
  int count() const { return array.count(); };
  SsSetElem *first() const { return (SsSetElem *)array.first(); };
  void reset(){ array.reset(); };
  void insert(void *data){
    array.insert(data,data);
  };
  int find(void *data) const { 
    return array.find(data)!=0;
  };
  void remove(const void *data){ array.remove(data); };
  int operator[](void *data) const { return find(data); };
  void xfer(SsSetOfString *pOther){ array.xfer(&pOther->array); };
};

#endif

/*
 * This routine removes all data elements from an Array.
 *
 * The data itself may or not be destroyed, 
 * depending upon the subclass of the Array.
 */
PRIVATE void SsArray::xClear(){
  SsArrayElem *pElem;
  SsArrayElem *pNextElem;

  for(pElem = pFirst; pElem; pElem = pNextElem){
    pNextElem = pElem->pNext;
    xDelete(pElem->pKey);
    xDeleteData(pElem->pData);
    delete pElem;
  }
  if( aHt ){
    delete [] aHt;
  }
  xReset();
}

/*
 * This routine initializes the private fields of an Array.
 */
PRIVATE void SsArray::xReset(){
  nElem = 0;
  pFirst = 0;
  nHtBucket = 0;
  aHt = 0;
}

/*
 * Change the size of the hash table to newSize.
 */
PRIVATE void SsArray::xRehash(int newSize){
  struct SsArray_ht *aNewHt;
  SsArrayElem *pList, *pNext;
  int i;

  if( newSize == nHtBucket ) return;
  aNewHt = new struct SsArray_ht[newSize];
  for(i=0; i<newSize; i++){
    aNewHt[i].nCollide = 0;
    aNewHt[i].pChain = 0;
  }
  if( aHt ) delete [] aHt;
  nHtBucket = newSize;
  aHt = aNewHt;
  pList = pFirst;
  pFirst = 0;
  while( pList ){
    pNext = pList->pNext;
    pList->pNext = 0;
    xLinkElem(pList,xHash(pList->pKey));
    pList = pNext;
  }
}

/*
 * Given an array element and its hash, insert the element into
 * the array at the appropriate place.
 */
PRIVATE void SsArray::xLinkElem(SsArrayElem *pElem, int h){
  h %= nHtBucket;
  if( aHt[h].nCollide==0 ){
    pElem->pPrev = 0;
    pElem->pNext = pFirst;
    if( pFirst ){
      pFirst->pPrev = pElem;
    }
    pFirst = pElem;
    aHt[h].pChain = pElem;
  }else{
    SsArrayElem *p = aHt[h].pChain;
    pElem->pPrev = p;
    pElem->pNext = p->pNext;
    if( p->pNext ){
      p->pNext->pPrev = pElem;
    }
    p->pNext = pElem;
  }
  aHt[h].nCollide++;
}

/*
 * Insert a new element into the array.  If there already exists another
 * element with the same key, then the data for that element is replaced
 * and the old data is returned.  If no element currently exists with the
 * same key, a new element is created and NULL is returned.
 *
 * Inserting NULL data is the same as removing the element.
 *
 * This method will not destroy the old data.  The calling function must
 * check the return of this method and destroy the data itself.
 */
PUBLIC void *SsArray::insert(void *key, void *data){
  SsArrayElem *pElem;
  int h, hOrig;
  int n;

  if( data==0 ){
    return remove(key);
  }
  if( nHtBucket<=0 ){
    xRehash(10);
  }
  hOrig = xHash(key);
  h = hOrig % nHtBucket;
  n = aHt[h].nCollide;
  pElem = aHt[h].pChain;
  while( n>0 && pElem ){
    if( xCompare(pElem->pKey,key)==0 ){
      void *pOldData = pElem->pData;
      pElem->pData = xCopyData(data);
      return pOldData;
    }
    n--;
    pElem = pElem->pNext;
  }
  pElem = new SsArrayElem();
  pElem->pKey = xCopy(key);
  pElem->pData = xCopyData(data);
  nElem++;
  if( nElem > nHtBucket * 2 ){
    xRehash(nHtBucket * 3);
  }
  xLinkElem(pElem,hOrig);
  return 0;  
}

/*
 * Locate an element with the given key and return a pointer to its
 * data.  Return NULL if nothing could be found.
 */
PUBLIC void *SsArray::find(void *key) const {
  SsArrayElem *pElem;
  int h;
  int n;

  if( nHtBucket<=0 ) return 0;
  h = xHash(key) % nHtBucket;
  n = aHt[h].nCollide;
  pElem = aHt[h].pChain;
  while( n>0 && pElem ){
    if( xCompare(pElem->pKey,key)==0 ){
      return pElem->pData;
    }
    n--;
    pElem = pElem->pNext;
  }
  return 0;
}

/*
 * Remove the element from the array that has the given key.  Return
 * the data associated with that element.  If no such element exists,
 * leave the array unaltered and return NULL.
 *
 * This method will not destroy the data.  The calling function must check
 * the return value and destroy it.
 */
PUBLIC void *SsArray::remove(const void *key){
  SsArrayElem *pElem;
  int h;
  int n;

  if( nHtBucket<=0 ) return 0;
  h = xHash(key) % nHtBucket;
  n = aHt[h].nCollide;
  pElem = aHt[h].pChain;
  while( n>0 && pElem ){
    if( xCompare(pElem->pKey,key)==0 ){
      void *pOldData;
      if( pElem->pPrev ){
        pElem->pPrev->pNext = pElem->pNext;
      }else{
        pFirst = pElem->pNext;
      }
      if( pElem->pNext ){
        pElem->pNext->pPrev = pElem->pPrev;
      }
      aHt[h].nCollide--;
      if( aHt[h].nCollide==0 ){
        aHt[h].pChain = 0;
      }else if( pElem==aHt[h].pChain ){
        aHt[h].pChain = pElem->pNext;
      }
      nElem--;
      pOldData = pElem->pData;
      xDelete(pElem->pKey);
      delete pElem;
      if( nHtBucket > 10 && nElem*10 < nHtBucket ){
        xRehash(nElem);
      }
      return pOldData;
    }
    n--;
    pElem = pElem->pNext;
  }
  return 0;
}

/*
** Transfer all the data out of one array into another.
*/
PUBLIC void SsArray::xfer(SsArray *pOther){
  pOther->nElem = nElem;
  pOther->pFirst = pFirst;
  pOther->nHtBucket = nHtBucket;
  pOther->aHt = aHt;
  xReset();
}

#ifdef ARRAY_TEST
#include <stdio.h>

int main(int argc, char **argv){
  SsArrayOfString x;
  char line[1000];

  while( fgets(line,sizeof(line),stdin) ){  
    int nArg;
    char arg0[100];
    char arg1[100];
    char arg2[100];
    if( line[0]=='#' ) continue;
    nArg = sscanf(line,"%[^ \t\n] %[^ \t\n] %[^ \t\n]",arg0,arg1,arg2);
    if( nArg==0 ) continue;
    printf("> %s",line);
    if( strcmp(arg0,"print")==0 ){
      SsArrayElem *i;
      for(i=x.first(); i; i = i->next()){
        printf("%s - %s\n",(char*)i->key(),(char*)i->data());
      }
    }else if( strcmp(arg0,"insert")==0 && nArg>=2 ){
      char *data;
      if( nArg==3 ){
        data = new char[strlen(arg2) + 1];
        strcpy(data,arg2);
      }else{
        data = 0;
      }
      void *old = x.insert(arg1,data);
      printf("old(%s) = %s\n",arg1,old ? (char*)old : "NULL");
      if( old ) delete [] old;
    }else if( strcmp(arg0,"find")==0 && nArg>=2 ){
      char *data = (char*)x.find((void*)arg1);
      printf("find(%s) = %s\n",arg1,data ? data : "NULL");
    }else if( strcmp(arg0,"count")==0 ){
      printf("count = %d\n",x.count());
    }else if( strcmp(arg0,"clear")==0 ){
      x.reset();
    }else{
      printf("Bad command: %s\n",arg0);
    }
  }
  return 0;
}
#endif
