/*
** This module contains methods that display the spreadsheet.
*/
static char const rcsid[] = "@(#) $Header: /home/softools/cvs/softools/code/common/cked/src/ss_display.cpp,v 1.79 2004/04/19 03:05:48 mark Exp $";

#include <tcl.h>
#include <tk.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <assert.h>
#include "ss_display.h"

#ifndef DEBUG
#undef assert
#define assert(expr)
#endif

/*
** Schedule the spreadsheet to be redisplayed at idle time,
** or right now if bNow is nonzero.
** If bTrim is nonzero, then the spreadsheet will trim its number
** of rows when it is redisplayed.
** If bAddRows is nonzero, then the spreadsheet will add rows to the
** bottom of the screen to fill empty space there.
*/
PUBLIC void ss::Display(int bNow, int bTrim, int bAddRows){
  if( bTrim ){
    displayProperties |= DISPLAY_TRIM_ROWS;
  }
  if( bAddRows ){
    displayProperties |= DISPLAY_ADD_ROWS;
  }

  /* Note that we ignore the bNow setting if there's already a pending
  ** spreadsheet redisplay.  So in that case, we wait until idle time anyway.
  */
  if( !(displayProperties & DISPLAY_PENDING) ){
    displayProperties |= DISPLAY_PENDING;
    if( bNow ){
      ss::Redisplay((ClientData) this);
    } else {
      Tcl_DoWhenIdle(ss::Redisplay, (ClientData) this);
    }
  }
}

PUBLIC void ss::Display(int bNow, int bTrim){
  Display(bNow, bTrim, 1);
}

/* By default, the redisplay is at idle time and no trimming is done.
*/
PUBLIC void ss::Display(){
  Display(0, 0, 1);
}

/*
** Cancel any currently scheduled spreadsheet redisplay.
*/
PUBLIC void ss::CancelRedisplay(){
  if( displayProperties & DISPLAY_PENDING ){
    Tcl_CancelIdleCall(ss::Redisplay, (ClientData) this);
  }
}

/*
** This method is an idle callback.  It invokes a Tcl command
** that will redisplay the spreadsheet.  (The Tcl command is useful to
** facilitate performance debugging.)
*/
PRIVATE STATIC void ss::Redisplay(ClientData clientData){
  ss *p = (ss *)clientData;
  char zThis[64];
  sprintf(zThis, "::ss_display %x", p);
  Tcl_Eval(p->interp, zThis);
}

/*
** This method implements the Tcl command
**
**  ::ss_display OBJECT_PTR
**
** where OBJECT_PTR is an ascii representation of the hex address of
** the "ss" object being displayed.
**
** The Tcl interface is used so that debugging tools that need a Tcl
** interface can report things like input/output parameters, tracebacks,
** and performance data.
*/
PUBLIC STATIC int ss::DisplayCmd(
  void *NotUsed,
  Tcl_Interp *interp,
  int argc,
  const char **argv
){
  if( argc!=2 ){
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
       " OBJECT_PTR\"", 0);
    return TCL_ERROR;
  }
  ss *p = (ss *)strtol(argv[1], 0, 16);
  if( !p ) return TCL_ERROR;
  p->ReallyRedisplay();
  return TCL_OK;
}

/*
** This method is an idle callback.  It redisplays the spreadsheet.
** The values from the spreadsheet's widget variable will be used to
** populate the spreadsheet cells.
**
** Redisplay utilizes "logical clipping".  Only the currently visible
** rows are displayed.  This is very fast, compared to displaying the
** entire spreadsheet.
*/
PRIVATE void ss::ReallyRedisplay(){
  displayProperties &= ~DISPLAY_PENDING;

  evalf("winfo exists %s", zSs);
  if( !evalIntResult() ) return;

  int pad = cellPad*2;

  /* Calculate the width and height of the viewport in pixels.
  ** If we can't calculate either of them, do nothing.
  */
  int width = CalculateViewportWidth();
  int height = CalculateViewportHeight();
  if( width<=0 || height<=0 ) return;

  fastcan.Eval("config -width %d -height %d", width, height);

  /* Save the active and text selection states.  They will be restored
  ** after redisplaying the cells.
  */
  int iCursor = fastcan.Index("active", "insert");
  SsString activeValue(GetWidgetVariable("active"));
  SsTclList *pTextSelection = GetCurrentTextSelection();

  /* Wipe out all of the state on the underlying canvas
  */
  fastcan.DeleteAll();
  fastcan.Eval("delete all");
  fastcan.Eval("select clear");
  fastcan.DeselectAll();

  /* Determine the number of rows to display.
  ** If we're not adding new rows, make sure the new number isn't larger than
  ** the old one.
  */
  int prevDisplayedRows = nDisplayedRows;
  nDisplayedRows = 
      GetHighestUndisplayedRow(displayProperties & DISPLAY_TRIM_ROWS);
  if( !(displayProperties & DISPLAY_ADD_ROWS) && 
      nDisplayedRows > prevDisplayedRows ){
    nDisplayedRows = prevDisplayedRows;
  }
  displayProperties &= ~DISPLAY_TRIM_ROWS;
  displayProperties &= ~DISPLAY_ADD_ROWS;

  /* If at least one cell has changed, or if we're adding new rows,
  ** calculate the row heights.
  */
  if( bDirtyCell || nDisplayedRows > prevDisplayedRows ){
    CalculateRowHeights(0, nDisplayedRows);
    bDirtyCell = 0;
  }

  /* Calculate all of the column widths once, before entering any loops.
  */
  int *pWidthsList = new int[sConfigure.cols];
  for( int col=0; col < sConfigure.cols; col++ ){
    pWidthsList[col] = GetColumnWidth(col);
  }
  
  /* Get the scrollable width and height, which is the viewport width and
  ** height, not counting the titles.  The titles don't get scrolled.
  */
  int scrollableWidth = width - pWidthsList[0];
  int scrollableHeight = height - GetConfiguredHeight(0);

  /* Get the new leftmost column and new top row, which might not
  ** be the same as the row and column specified by "topleft":
  ** They can change if the canvas widget got resized.
  */
  int rightmostLeftcol = 
      GetColWithinDistance(sConfigure.cols, -scrollableWidth);
  int newLeftcol = min(topleft.GetCol(), rightmostLeftcol);
  int lastCol = GetColAtDistance(newLeftcol, scrollableWidth);
  int shiftX = GetDistanceBetweenCols(1, newLeftcol);

  int lowestToprow = GetRowWithinDistance(nDisplayedRows, -scrollableHeight);
  int newToprow = min(topleft.GetRow(), lowestToprow);
  int lastRow = GetRowAtDistance(newToprow, scrollableHeight);
  if( lastRow>=nDisplayedRows ){
    lastRow = nDisplayedRows - 1;
  }
  int shiftY = GetDistanceBetweenRows(1, newToprow);

  /* Initialize the commands that will be used to create cells in the 
  ** display loop below.  Initializing these commands once outside of the
  ** loop saves at least 2/3 of the processing time.
  */
  char *zLastFill = sConfigure.zBg;
  SsCellCommand *pCellCommand = 
      fastcan.CreateCellCommand("notresize cell", zLastFill);
  char *zLastTextFill = sConfigure.zFg;
  SsTextCommand *pTextCommand = fastcan.CreateTextCommand(sConfigure.zFont, 
      "notresize text", "w", zLastTextFill);

  /* Display the non-title cells.
  ** Each cell is a canvas rectangle.
  */
  register int y = GetConfiguredHeight(0) + shiftY;
  register int x = 0;
  char zCellTag[67];
  zCellTag[0] = 'C';
  char *zCell = zCellTag+1;
  assert(sConfigure.cols<100);
  for( register int row=newToprow; row <= lastRow; row++ ){
    int *pWidths = pWidthsList;
    x = *pWidths++;
    int rowHeight = GetConfiguredHeight(row);
    int y1 = y + rowHeight - 1;
    sprintf(zCell, "%d,", row);
    char *zCellCol = zCell + strlen(zCell);
    zCellCol[0] = '0';
    zCellCol[1] = 0;
    pCellCommand->SetYCoords(y, y1);
    pTextCommand->SetYCoord(y + rowHeight/2);
 
    /* Note well:  The following code must be made as fast as possible,
    ** since it is executed once for each cell in the viewable portion
    ** of the spreadsheet.
    */
    for( register int col=1; col < sConfigure.cols; col++ ){
      if( col<10 ){
        zCellCol[0]++;
      } else {
        zCellCol[0] = '0' + col/10;
        zCellCol[1] = '0' + col%10;
        zCellCol[2] = 0;
      }
      int colWidth = *pWidths++;
      if( colWidth<=0 ) continue;
      int nextX = x + colWidth;

      char *zCellFill = (char *)aBg[(void *)zCell];
      if( !zCellFill || !*zCellFill ){
        zCellFill = sConfigure.zBg;
      }
      if( zCellFill!=zLastFill ){
        pCellCommand->SetFill(zCellFill);
        zLastFill = zCellFill;
      }
      pCellCommand->SetXCoords(x, nextX-1);
      int ret = pCellCommand->Execute();
      int id = 0;
      Tcl_GetIntFromObj(0, Tcl_GetObjResult(interp), &id);
      assert(id>0);
      fastcan.AddOneToOneTag(zCellTag, id);
      fastcan.MapIdToCell(id, row, col);

      /* Don't create the text glyph now if there is no data to display.
      ** It will be created "on demand" later.  This makes redisplay go much
      ** faster for large spreadsheets.
      */
      const char *zText = GetWidgetVariable(zCell);
      if( zText && *zText && colWidth>pad ){
        char *zTextFill = (char *)aFg[(void *)zCell];
        if( !zTextFill || !*zTextFill ){
          zTextFill = sConfigure.zFg;
        }
        if( zTextFill!=zLastTextFill ){
          pTextCommand->SetFill(zTextFill);
          zLastTextFill = zTextFill;
        }
        pTextCommand->SetXCoord(x+cellPad);
        pTextCommand->SetText(zText);
        pTextCommand->SetWidth(colWidth - pad);
        int ret = pTextCommand->Execute();
        int textId = 0;
        Tcl_GetIntFromObj(0, Tcl_GetObjResult(interp), &textId);
        assert(textId > 0);
        fastcan.AddOneToOneTag(zCell, textId);
        fastcan.MapIdToCell(textId, row, col);
      }
      x = nextX;
    }
    y = y1 + 1;
  }
  int lastX = x;
  int lastY = y;
  delete pCellCommand;
  delete pTextCommand;

  /* Draw the watermark, if there is one.
  */
  if( sConfigure.zWaterImage && *sConfigure.zWaterImage ){
    DrawWatermark();
  }

  /* Now draw stippled lines that indicate the hidden columns, if there
  ** are any.
  */
  SsLineCommand *pHideCommand = 
      fastcan.CreateLineCommand("vhidden", sConfigure.zHideColor);
  pHideCommand->SetYCoords(sConfigure.titleRowHeight, lastY);
  pHideCommand->SetStipple(sConfigure.zHideColor);
  pHideCommand->SetWidth(6);
  int stippling = 0;
  x = pWidthsList[0];
  for( int col=1; col < sConfigure.cols; col++ ){
    int colWidth = pWidthsList[col];
    if( colWidth<=0 ) {  // column is hidden
      ++stippling;
      continue;
    }
    if( stippling ){
      pHideCommand->SetXCoords(x, x);
      pHideCommand->Execute();
      stippling = 0;
    }
    x += colWidth;
  }

  /* Now that all "normal" cells have been displayed, display the title cells.
  ** Order is important:  The title cells should always be on top of normal
  ** cells, so that scrolling the normal cells won't overwrite the title cells.
  ** So we display the title cells after the normal cells to make the
  ** stacking order right.
  **
  ** Doing it this way is faster than issuing a bunch of "raise" commands.
  **
  ** First, the title row:
  */
  SsCellCommand *pCellTitleCommand = 
      fastcan.CreateCellCommand("notresize title", sConfigure.zTitleBg);
  SsTextCommand *pTextTitleCommand = fastcan.CreateTextCommand(sConfigure.zFont,
      "notresize title", "c", sConfigure.zTitleFg);
  x = pWidthsList[0];
  y = 0;
  int y1 = GetConfiguredHeight(0) - 1;
  int nextY = y1 + 1;
  int textY = nextY/2;

  for( int col=1; col < sConfigure.cols; col++ ){
    int colWidth = pWidthsList[col];
    if( colWidth<=0 ) continue;

    char zCellTag[67];
    sprintf(zCellTag, "C0,%d", col);
    char *zCell = zCellTag+1;

    int nextX = x + colWidth;

    pCellTitleCommand->SetCoords(x, 0, nextX-1, nextY);
    int ret = pCellTitleCommand->Execute();
    int id = evalIntResult();
    assert(ret==TCL_OK && id > 0);
    fastcan.AddTag(zCellTag, id);
    fastcan.AddTag("titlecell", id);
    fastcan.AddTag("my", id);
    fastcan.MapIdToCell(id, 0, col);

    if( colWidth>pad ){
      const char *zText = GetWidgetVariable(zCell);
      if( !zText ) zText = "";
      int textX = x + colWidth/2;

      pTextTitleCommand->SetCoords(textX, textY);
      pTextTitleCommand->SetText(zText);
      pTextTitleCommand->SetWidth(colWidth - pad);
      ret = pTextTitleCommand->Execute();
      int textId = evalIntResult();
      assert(ret==TCL_OK && textId>0);
      fastcan.AddTag(zCell, textId);
      fastcan.AddTag("title", textId);
      fastcan.AddTag("my", textId);
      fastcan.MapIdToCell(textId, 0, col);
    }

    x = nextX;
  }

  x = 0;
  y = y1 + 1 + shiftY;
  int colWidth = pWidthsList[0];
  if( colWidth>0 ){

    /* Now the title column:
    */
    if( colWidth>pad ){
      pTextTitleCommand->SetWidth(colWidth - pad);
      pCellTitleCommand->SetXCoords(0, colWidth);
      pTextTitleCommand->SetXCoord(colWidth/2);
    }
    for( register int row=newToprow; row <= lastRow; row++ ){
      char zCellTag[67];
      sprintf(zCellTag, "C%d,0", row);
      char *zCell = zCellTag+1;
   
      int rowHeight = GetConfiguredHeight(row);
      int textY = y + rowHeight/2;
      int nextY = y + rowHeight;
      int ret;

      pCellTitleCommand->SetYCoords(y, nextY);
      ret = pCellTitleCommand->Execute();
      int id = 0;
      Tcl_GetIntFromObj(0, Tcl_GetObjResult(interp), &id);
      assert(id > 0);
      fastcan.AddTag(zCellTag, id);
      fastcan.AddTag("titlecell", id);
      fastcan.AddTag("mx", id);
      fastcan.MapIdToCell(id, row, 0);

      const char *zText = GetWidgetVariable(zCell);
      if( !zText ) zText = "";
      if( colWidth>pad ){
        pTextTitleCommand->SetYCoord(textY);
        pTextTitleCommand->SetText(zText);
        ret = pTextTitleCommand->Execute();
        int textId = 0;
        Tcl_GetIntFromObj(0, Tcl_GetObjResult(interp), &textId);
        assert(textId>0);
        fastcan.AddTag(zCell, textId);
        fastcan.AddTag("title", textId);
        fastcan.AddTag("mx", textId);
        fastcan.MapIdToCell(textId, row, 0);
      }

      y = nextY;
    }

    /* The upperleft cell comes next.
    */
    {
      x = 0;
      y = 0;
      int rowHeight = GetConfiguredHeight(0);
      int y1 = rowHeight - 1;
      int x1 = colWidth - 1;

      int nextX = x1;
      int nextY = y1;

      pCellTitleCommand->SetCoords(x, y, nextX, nextY);
      int ret = pCellTitleCommand->Execute();
      int id = evalIntResult();
      assert(ret==TCL_OK && id > 0);
      fastcan.AddTag("C0,0", id);
      fastcan.AddTag("upperleft", id);
      fastcan.AddTag("titlecell", id);
      fastcan.AddTag("mx", id);
      fastcan.AddTag("my", id);
      fastcan.MapIdToCell(id, 0, 0);

      if( colWidth>pad ){
        const char *zText = GetWidgetVariable("0,0");
        if( !zText ) zText = "";
        pTextTitleCommand->SetCoords(colWidth/2, rowHeight/2);
        pTextTitleCommand->SetText(zText);
        pTextTitleCommand->SetWidth(colWidth - pad);
        ret = pTextTitleCommand->Execute();
        int textId = evalIntResult();
        assert(ret==TCL_OK && textId>0);
        fastcan.AddTag("0,0", textId);
        fastcan.AddTag("title", textId);
        fastcan.AddTag("mx", textId);
        fastcan.AddTag("my", textId);
        fastcan.MapIdToCell(textId, 0, 0);
      }
    }
  }

  /* Next, draw the lines that allow the user to resize rows and columns.
  ** Order is important:  Drawing this after cells avoids having these lines
  ** become obscured by any of the cells drawn previously.
  */
  DrawResizers(newToprow, lastRow, shiftY);

  /* All done with cell displays!  Clean up.
  */
  delete [] pWidthsList;
  delete pCellTitleCommand;
  delete pTextTitleCommand;

  /* Draw a box under the cells, which is a trick that allows 
  ** us to avoid drawing a bunch of border lines.
  */
  if( sConfigure.zBdColor && *sConfigure.zBdColor ){
    fastcan.Eval("create rectangle %d %d %d %d -fill %s -outline {}"
        " -tags notresize",
        0, 0, lastX, lastY, sConfigure.zBdColor);
    int id = evalIntResult();
    assert(id>0);
    fastcan.AddTag("bgrect", id);
    fastcan.Eval("lower %d", id);
  }

  /* Draw the scrollable region.
  */
  int canWidth = GetDistanceBetweenCols(1, rightmostLeftcol) + width + 1;
  int canHeight = GetDistanceBetweenRows(1, lowestToprow) + height + 1;
  fastcan.Eval("config -scrollregion \"0 0 %d %d\" "
      "-xscrollincrement 1 -yscrollincrement 1",
      canWidth, canHeight);

  /* Restore the active state that was saved previously.
  ** But watch out for the case where the active cell has disappeared from
  ** the displayed spreadsheet.
  */
  if( !activeCell.isnull() ){
    int activeRow = activeCell.GetRow();
    int activeCol = activeCell.GetCol();
    if( activeRow<0 || activeRow>=nDisplayedRows ||
        activeCol<0 || activeCol>=sConfigure.cols ){
      activeCell.clear();
      SetWidgetVariable("active", "");
    } else {
      char *zActiveCell = activeCell.render();
      char *zActiveValue = activeValue.render();
      SsString cellTag("C"); cellTag += zActiveCell;
      fastcan.AddTag("activecell", cellTag.render());
      fastcan.SetText(zActiveCell, zActiveValue, this);
      fastcan.AddTag("active", zActiveCell);
      fastcan.Focus(zActiveCell);
      if( iCursor!=-1 ){
        fastcan.Icursor(zActiveCell, iCursor);
      }
    }
  }
  DrawSelectionRectangle();

  /* Color the selected cells and the active cell.
  */
  ResyncSelection();
  ColorAllSelected();
  if( pTextSelection ){
    SetTextSelection(pTextSelection->get(0), 
        pTextSelection->get_int(1), pTextSelection->get_int(2));
    delete pTextSelection;
  }

  /* Shift the unscrollable canvas items based on the pixel location 
  ** of the topleft row.
  */
  if( shiftX!=0 ){
    fastcan.Move("mx", shiftX, 0);
  }
  if( shiftY!=0 ){
    fastcan.Move("my", 0, shiftY);
  }
  if( canWidth > 0 ){
    double fract = (double)shiftX/(double)canWidth;
    fastcan.Eval("xview moveto %g", fract);
  }
  if( canHeight > 0 ){
    double fract = (double)shiftY/(double)canHeight;
    fastcan.Eval("yview moveto %g", fract);
  }
  topleft.set(newToprow, newLeftcol);
}

/*
** This routine will draw a watermark on the spreadsheet.
** It will be set up so that it won't appear to move when
** the spreadsheet is scrolled.
*/
PRIVATE void ss::DrawWatermark(){
  fastcan.Eval("cget -width");
  int cx = evalIntResult()/2;
  fastcan.Eval("cget -height");
  int cy = evalIntResult()/2;
  fastcan.Eval("create image %d %d -anchor c -image %s -tags {notresize wm}",
       cx, cy, sConfigure.zWaterImage);
  int wid = evalIntResult();
  fastcan.AddTag("mx", wid);
  fastcan.AddTag("my", wid);
  fastcan.Eval("raise text");
  fastcan.Eval("lower %d text", wid);
}

/*
** This routine will draw lines that will allow the user to resize
** rows and columns in the spreadsheet.
*/
PRIVATE void ss::DrawResizers(int topRow, int lastRow, int shiftY){
  int titleRowHeight = GetConfiguredHeight(0);
  int y = titleRowHeight - 1 + shiftY;
  int x = GetColumnWidth(0) - 1;

  /* Create horizontal resizers.
  */
  SsLineCommand *pLineCommand = 
      fastcan.CreateLineCommand("hresize", sConfigure.zResizeColor);
  pLineCommand->SetXCoords(0, x);
  for( int row=topRow; row <= lastRow+1; row++ ){

    /* Draw the horizontal line that will be used to resize the row.
    ** The row tag references the row *below* the horizontal resizer;
    ** e.g. "RR20" is the tag of the resizer just above row 20.
    ** ("RR" stands for "resize row".)
    */
    if( row==1 ){
      SsLineCommand *pTopLineCommand = 
          fastcan.CreateLineCommand("title", sConfigure.zResizeColor);
      pTopLineCommand->SetCoords(0, y, x, y);
      pTopLineCommand->Execute();
      delete pTopLineCommand;
    } else {
      pLineCommand->SetYCoords(y, y);
      pLineCommand->Execute();
    }
    int id = evalIntResult();
    assert(id > 0);
    fastcan.AddTag("mx", id);
    fastcan.AddTag("resize", id);
    SsString tag("RR");
    tag += row;
    fastcan.AddTag(tag.render(), id);

    y += GetConfiguredHeight(row);
  }
  delete pLineCommand;

  /* Create vertical resizers.
  ** First, create the Tcl commands that will be used to draw them.
  ** Then do the actual drawing.
  */
  SsLineCommand *pColCommand = 
      fastcan.CreateLineCommand("vresize", sConfigure.zResizeColor);
  int yTop = titleRowHeight;
  pColCommand->SetYCoords(0, yTop);

  int stippling = 0;
  x = GetColumnWidth(0);
  for( int col=1; col <= sConfigure.cols; col++ ){
    int colWidth = GetColumnWidth(col);
    if( colWidth<=0 ){  // The next resizer after a hidden column is stippled.
      ++stippling;
      continue;
    }

    /* Draw the vertical line that will be used to resize the column.
    ** The column tag contains the column to the *right* of the resizer;
    ** e.g. "RC5" is the tag of the resizer to the left of column 5.
    ** ("RC" stands for "resize column".)
    */
    if( col==1 ){
      SsLineCommand *pLeftColCommand = 
          fastcan.CreateLineCommand("title vresize", sConfigure.zResizeColor);
      pLeftColCommand->SetCoords(x, 0, x, yTop);
      pLeftColCommand->Execute();
      delete pLeftColCommand;
    } else {
      if( stippling ){
        pColCommand->SetStipple(sConfigure.zHideColor);
        pColCommand->SetWidth(6);
      } else {
        pColCommand->SetStipple("");
        pColCommand->SetWidth(0);
      }
      pColCommand->SetXCoords(x, x);
      pColCommand->Execute();
    }
    int id = evalIntResult();
    assert(id > 0);
    fastcan.AddTag("my", id);
    fastcan.AddTag("resize", id);
    SsString tag("RC");
    tag.append(col);
    fastcan.AddTag(tag.render(), id);
    stippling = 0;

    x += colWidth;
  }
  delete pColCommand;
}

/*
** This method will copy all the data within the range of rows given
** by "from" and "to", onto the canvas named zNewCanvas.
** "rowMag" is the magnification for every row.
*/
PRIVATE void ss::CopySpreadsheet(int from, int to, char *zNewCanvas, 
    float rowMag){
  int pad = cellPad*2;

  fastcan.Eval("cget -bg");
  evalf("%s config -bg %s", zNewCanvas, evalStrResult());

  /* Clear the canvas
  */
  evalf("%s delete all", zNewCanvas);
  evalf("%s select clear", zNewCanvas);

  /* Create horizontal borders.
  **
  ** Note that we must create borders in order for them to be visible;
  ** we can't use the trick of creating a large rectangle under all of the
  ** cells.  This is because the canvas is going to be rendered in postscript,
  ** and postscript could easily cause the borders to disappear because of
  ** scaling.
  */
  int y = GetConfiguredHeight(0) - 1;
  int yBot = y;
  int x = GetColumnWidth(0) - 1;
  int xRight = x + GetDistanceBetweenCols(1, sConfigure.cols);

  int row = 0;
  while( row <= to ){
    evalf("%s create line 0 %d %d %d -fill %s",
        zNewCanvas, y, x, y, sConfigure.zResizeColor);
    int id = evalIntResult();
    assert(id > 0);

    if( row==0 ){
      evalf("%s itemconfig %d -tags {title resize}", zNewCanvas, id);
    } else {
      evalf("%s itemconfig %d -tags resize", zNewCanvas, id);
    }

    evalf("%s create line %d %d %d %d -fill %s -tags border",
        zNewCanvas, x, y, xRight, y, sConfigure.zBdColor);
    id = evalIntResult();
    assert(id > 0);

    yBot = y;
    if( row==0 && from>0 ){
      row = from;
    } else {
      row++;
    }
    y += (int)(GetConfiguredHeight(row) * rowMag);
  }

  /* Create vertical borders.
  */
  int yTop = GetConfiguredHeight(0);
  yBot++;
  int stippling = 0;
  for( int col=1; col <= sConfigure.cols; col++ ){
    int colWidth = GetColumnWidth(col);
    if( colWidth<=0 ){
      stippling = 1;
      continue;
    }
    evalf("%s create line %d 0 %d %d -fill %s -tags resize",
        zNewCanvas, x, x, yTop, sConfigure.zResizeColor);
    int id = evalIntResult();
    assert(id > 0);
    if( col==1 ){
      evalf("%s addtag title withtag %d", zNewCanvas, id);
    }
    if( stippling ){
      evalf("%s addtag vhidden withtag %d", zNewCanvas, id);
      evalf("%s itemconfig %d -width 6 -stipple %s", 
          zNewCanvas, id, sConfigure.zHideColor);
    }

    evalf("%s create line %d %d %d %d -fill %s -tags border",
          zNewCanvas, x, yTop, x, yBot, sConfigure.zBdColor);
    id = evalIntResult();
    assert(id > 0);
    if( stippling ){
      evalf("%s addtag vhidden withtag %d\n", zNewCanvas, id);
      evalf("%s itemconfig %d -width 6 -stipple %s",
          zNewCanvas, id, sConfigure.zHideColor);
    }

    x += colWidth;
    stippling = 0;
  }
  
  /* Create the title cells and the data cells.
  */
  y = 0;
  row = 0;
  SsCell *pActive = Index("active");
  while( row <= to ){
    x = 0;
    int rowHeight = GetConfiguredHeight(row);
    if( row!=0 ){
      rowHeight = (int)(rowHeight * rowMag);
    }
    int y1 = y + rowHeight - 1;
    int halfRowHeight = rowHeight/2;
 
    for( int col=0; col < sConfigure.cols; col++ ){
      int colWidth = GetColumnWidth(col);
      if( colWidth<0 ) continue;
      int x1 = x + colWidth - 1;
      int halfColWidth = colWidth/2;

      char *zAnchor = "c";
      int textX, textY;
      int nextX, nextY;
      SsString cTags;

      SsCell cell(row, col);
      char *zCell = cell.render();

      if( row==0 && col==0 ){
        textX = colWidth/2;
        textY = halfRowHeight;
        nextX = x1;
        nextY = y1;
        cTags.set("upperleft titlecell");
      } else if( row==0 ){
        textX = x + halfColWidth;
        textY = halfRowHeight;
        nextX = x1 + 1;
        nextY = y1 + 1;
        cTags.set("titlecell");
      } else if( col==0 ){
        textX = colWidth/2;
        textY = y + halfRowHeight;
        nextX = x1 + 1;
        nextY = y1 + 1;
        cTags.set("titlecell");
      } else {
        zAnchor = "w";
        textX = x + cellPad;
        textY = y + halfRowHeight;
        nextX = x1;
        nextY = y1;
      }

      /* Create the cell.
      */
      evalf("%s create rectangle %d %d %d %d -outline {} -tags {cell C%s}",
          zNewCanvas, x, y, nextX, nextY, zCell);
      int id = evalIntResult();
      assert(id > 0);

      for( char *zTag = cTags.first(); zTag; zTag = cTags.next() ){
        evalf("%s addtag %s withtag %d", zNewCanvas, zTag, id);
      }
      if( pActive && pActive->equal(cell) ){
        evalf("%s addtag activecell withtag %d", zNewCanvas, id);
      }

      /* Create the text within the cell.
      */
      const char *zText = GetWidgetVariable(zCell);
      if( zText && *zText && colWidth>pad ){
        evalf("%s create text %d %d -anchor %s -font \"%q\" -text \"%q\" "
            "-width %d -tags %s",
            zNewCanvas, textX, textY, zAnchor, sConfigure.zFont, 
            zText, colWidth - pad, zCell);

        int textId = evalIntResult();
        if( row==0 || col==0 ){
          evalf("%s addtag title withtag %d", zNewCanvas, textId);
        } else {
          evalf("%s addtag text withtag %d", zNewCanvas, textId);
        }
        if( pActive && pActive->equal(cell) ){
          evalf("%s addtag active withtag %d", zNewCanvas, textId);
        }
      }

      x = x1 + 1;
    }

    y = y1 + 1;

    if( row==0 && from>0 ){
      row = from;
    } else {
      row++;
    }
  }
  if( pActive) {
    delete pActive;
  }

  if( sConfigure.zWaterImage && *sConfigure.zWaterImage ){
    int cx = x/2;
    int cy = y/2;
    evalf("%s create image %d %d -anchor c -image %s -tags wm",
         zNewCanvas, cx, cy, sConfigure.zWaterImage);
    evalf("%s raise text", zNewCanvas);
    evalf("%s lower wm text", zNewCanvas);
  }

  /* Configure the size of the canvas.
  */
  evalf("%s config -width %d -height %d", zNewCanvas, x, y);

  /* Set tags for the selected cells.
  */
  SsTclList *pSelected = GetCurrentSelection();
  for( char *zCell = pSelected->first(); zCell; zCell = pSelected->next() ){
    evalf("%s addtag selectedcell withtag C%s", zNewCanvas, zCell);
    evalf("%s addtag selected withtag %s", zNewCanvas, zCell);
  }
  delete pSelected;

  /* Make sure the display list is ordered properly so that everything
  ** is displayed.
  */
  evalf("%s raise cell", zNewCanvas);
  evalf("%s raise border", zNewCanvas);
  evalf("%s raise text", zNewCanvas);
  evalf("%s raise titlecell", zNewCanvas);
  evalf("%s raise title", zNewCanvas);
  evalf("%s raise resize", zNewCanvas);
  evalf("%s raise upperleft", zNewCanvas);
  evalf("%s raise vhidden", zNewCanvas);
  evalf("%s raise wm", zNewCanvas);

  /* Configure the cloned canvas' background colors.
  */
  if( sConfigure.zBg && *sConfigure.zBg ){
    evalf("%s itemconfigure cell -fill %s", zNewCanvas, sConfigure.zBg);
  }
  for( SsArrayElem *pElem = aBg.first(); pElem; pElem = pElem->next() ){
    char *zCellBg = (char *)(pElem->data());
    if( *zCellBg ){
      evalf("%s itemconfigure C%s -fill %s", 
          zNewCanvas, (char *)(pElem->key()), zCellBg);
    }
  }
  if( sConfigure.zSelectBg && *sConfigure.zSelectBg ){
    evalf("%s itemconfigure selectedcell -fill %s",
        zNewCanvas, sConfigure.zSelectBg);
  }
  if( sConfigure.zActiveBg && *sConfigure.zActiveBg ){
    evalf("%s itemconfigure activecell -fill %s",
        zNewCanvas, sConfigure.zActiveBg);
  }
  if( sConfigure.zTitleBg && *sConfigure.zTitleBg ){
    evalf("%s itemconfigure titlecell -fill %s",
        zNewCanvas, sConfigure.zTitleBg);
  }

  /* Configure the cloned canvas' foreground colors.
  */
  if( sConfigure.zFg && *sConfigure.zFg ){
    evalf("%s itemconfigure text -fill %s", zNewCanvas, sConfigure.zFg);
  }
  for( SsArrayElem *pElem = aFg.first(); pElem; pElem = pElem->next() ){
    char *zCellFg = (char *)(pElem->data());
    if( *zCellFg ){
      evalf("%s itemconfigure %s -fill %s",
          zNewCanvas, (char *)(pElem->key()), zCellFg);
    }
  }
  if( sConfigure.zSelectFg && *sConfigure.zSelectFg ){
    fastcan.Itemconfigure("selected", "-fill", sConfigure.zSelectFg);
    evalf("%s itemconfigure selected -fill %s",
        zNewCanvas, sConfigure.zSelectFg);
  }
  if( sConfigure.zActiveFg && *sConfigure.zActiveFg ){
    evalf("%s itemconfigure active -fill %s", zNewCanvas, sConfigure.zActiveFg);
  }
  if( sConfigure.zTitleFg && *sConfigure.zTitleFg ){
    evalf("%s itemconfigure title -fill %s", zNewCanvas, sConfigure.zTitleFg);
  }
}

/*
** Trim rows off the bottom of the spreadsheet widget, such that:
** - no trimmed row has any data in it;
** - no row above the row specified by "-rows" is trimmed; and
** - no visible row is trimmed.
**
** This method may change the value of nDisplayedRows.
*/
PRIVATE void ss::TrimRows(){
  int firstTrimmableRow = GetHighestUndisplayedRow(/* bTrim== */ 1);
  if( firstTrimmableRow >= nDisplayedRows ){  // trimming not necessary
    return;
  }

  /* There is at least one row to trim.
  ** First, remove horizontal resizers and cells below the first trimmable row.
  */
  int x0 = -1;
  int xRight = GetColumnWidth(0) +
      GetDistanceBetweenCols(1, sConfigure.cols) + 1;
  int yTrimTop = GetRowY(firstTrimmableRow) - 1;
  int yTrimBot = GetRowY(nDisplayedRows) + 1;
  if( yTrimTop<0 || yTrimBot==0 ) return;
  fastcan.Eval("find enclosed %d %d %d %d", x0, yTrimTop, xRight, yTrimBot);
  const char *zIdList = evalStrResult();
  if( zIdList && *zIdList ){
    char *zCopy = new char[strlen(zIdList) + 1];  // since zIdList is volatile
    strcpy(zCopy, zIdList);
    fastcan.DeleteIds(zCopy);
    delete [] zCopy;
  }

  int id=fastcan.FindFirst("bgrect");
  if( id!=0 ){
    fastcan.Eval("coords %d %d %d %d %d", id, 0, 0, xRight-1, yTrimTop+1);
  }

  /* Make sure that any persistent state does not refer to the cells being 
  ** removed.
  */
  for( int row = firstTrimmableRow; row < nDisplayedRows; row++ ){
    for( int col = 0; col < sConfigure.cols; col++ ){
      SsCell cell(row, col);
      aBg.remove((void *)cell.render());
      aFg.remove((void *)cell.render());
      aHeight.remove((void *)row);
    }
  }
  if( selectionAnchor.GetRow() >= firstTrimmableRow ){
    selectionAnchor.set((char *)0);
  }

  /* Reset the widget variable's "heights" element.
  */
  if( sConfigure.zVariable && *sConfigure.zVariable ){
    SsTclList *pHeights = GetAllHeights();
    SetWidgetVariable("heights", pHeights->render());
    delete pHeights;
  }

  /* Reset the canvas scroll region.
  */
  int lowestToprow = GetRowWithinDistance(firstTrimmableRow, 
      -GetScrollableHeight());
  fastcan.Eval("cget -height");
  int height = evalIntResult();
  int canHeight = GetDistanceBetweenRows(1, lowestToprow) + height + 1;
  fastcan.Eval("cget -scrollregion");
  SsBbox bbox(evalStrResult());
  fastcan.Eval("config -scrollregion \"0 0 %d %d\"", bbox.x1, canHeight);
  nDisplayedRows = firstTrimmableRow;

  /* Reset the "selected" variable based on the newly trimmed spreadsheet.
  ** This must be done after nDisplayedRows is set.
  */
  ResyncSelection();
}

/* 
** Reset the "selected" variable and the canvas selection,
** taking into account the maximum number of displayed rows in nDisplayedRows.
**
** This can be called either when the maximum number of displayed rows changes,
** or when the "selected" widget variable value changes and it is necessary
** to sync its value up with the equivalent "fastcan" selection.
*/
PRIVATE void ss::ResyncSelection(){

  /* Shortcut case for selection value "all".
  */
  const char *zSelection = GetWidgetVariable("selected");
  if( zSelection && *zSelection=='a' && strcmp(zSelection, "all")==0 ){
    fastcan.DeselectAll();
    fastcan.SelectAll(nDisplayedRows, sConfigure.cols);
    return;
  }

  SsTclList *pSelected = GetCurrentSelection();
  fastcan.DeselectAll();
  SsTclList new_selected(interp);
  for( char *zCell = pSelected->first(); zCell; zCell = pSelected->next() ){
    int nrow, ncol;
    if( sscanf(zCell, "%d,%d", &nrow, &ncol)==2 && nrow<nDisplayedRows ){
      new_selected.append(zCell);
      fastcan.SelectCell(zCell);
    }
  }
  delete pSelected;
  SetSelectedVariable(&new_selected);
}

/*
** This method will draw a "selection rectangle."
** It is called whenever the selection changes.  It will draw
** a black rectangular border around the selection.
**
** The rectangle will only be drawn if the selection consists of a set of cells
** that form a rectangle.  If not, then any existing selection rectangle
** will be deleted.
*/
PUBLIC void ss::DrawSelectionRectangle(){
  fastcan.Eval("delete selrect");
  delete pSelectionRect;
  pSelectionRect = new SsTclList(interp);
  SsCell tlcell;
  SsCell brcell;
  if( !GetSelectionRectangle(tlcell, brcell) ){
    return;
  }
  if( !DrawRectangle(tlcell, brcell, 2, "black", "selrect") ){
    return;
  }
  pSelectionRect->append(tlcell.GetRow());
  pSelectionRect->append(tlcell.GetCol());
  pSelectionRect->append(brcell.GetRow());
  pSelectionRect->append(brcell.GetCol());
}

/*
** This method will return the selection rectangle's "bounding box".
** The bounding box consists of the rectangle's top row, left column,
** bottom row, and right column.
*/
PUBLIC char *ss::GetSelrectBbox(){
  return pSelectionRect->render();
}

/* This method looks at the current selection and determines whether it forms
** a "perfect rectangle"; that is, whether every cell in the rectangle 
** that it sweeps out is contained in the selection.
**
** If the selection forms a perfect rectangle, the method will return the
** top left cell and the bottom right cell in tlcell and brcell, and
** will return TRUE.
**
** If the selection isn't a perfect rectangle, the values in the cells are
** undefined, and the method will return FALSE.
*/
PRIVATE int ss::GetSelectionRectangle(SsCell &tlcell, SsCell &brcell){

  /* Shortcut case for the "all" selection.
  */
  const char *zSelection = GetWidgetVariable("selected");
  if( zSelection && *zSelection=='a' && strcmp(zSelection, "all")==0 ){
    tlcell.set(1, 1);
    brcell.set(nDisplayedRows-1, sConfigure.cols-1);
    return 1;
  }

  SsTclList *pSelection = GetCurrentSelection();
  char *zCell = pSelection->first();
  if( !zCell ){
    delete pSelection;
    return 0;
  }
  tlcell.set(zCell);
  brcell.set(zCell);
  int right = -1;                 // rightmost column
  int left = tlcell.GetCol();     // leftmost column
  int prevrow = tlcell.GetRow();
  int prevcol = tlcell.GetCol();

  /* For each cell in the selection (which is in row-major order),
  ** determine if the current cell would form a non-rectangular set of cells,
  ** taken with the cells visited so far.
  */
  for(zCell = pSelection->next() ; zCell; zCell = pSelection->next() ){
    brcell.set(zCell);
    if( brcell.GetRow()==prevrow ){
      if( brcell.GetCol()!=prevcol+1 || 
          (right!=-1 && brcell.GetCol()>right) ) break;
      prevcol++;
    } else {
      if( brcell.GetRow()!=prevrow+1 || brcell.GetCol()!=left ||
          (right!=-1 && prevcol!=right) ) break;
      if( right==-1 ) right = prevcol;
      prevrow++;
      prevcol = left;
    }
  }
  delete pSelection;
  int ret = (!zCell && (right==-1 || brcell.GetCol()==right));
  return ret;
}

/*
** This method will draw a rectangle on the spreadsheet.  The rectangle
** will encompass a rectangular set of cells on the spreadsheet.  The
** rectangle is formed by the top left cell in tlcell and the
** bottom right cell in brcell.
** The rectangle will be drawn with the given width, color, and tag.
**
** The rectangle will be clipped to the visible part of the spreadsheet,
** so that it won't overlap any title cells or other graphics outside
** of the visible spreadsheet.
**
** The method returns TRUE if the rectangle was actually drawn.
*/
PUBLIC int ss::DrawRectangle(
  SsCell &tlcell,
  SsCell &brcell,
  int width,
  const char *zColor,
  const char *zTag
){
  /* Find out the minimum and maximum row, and minimum column,
  ** for the rectangle.  The drawing will be clipped to these.
  */
  int topClip = topleft.GetRow();
  int leftClip = topleft.GetCol();
  int bottomClip = GetRowAtDistance(topleft.GetRow(), GetScrollableHeight());
  if( bottomClip>=nDisplayedRows ){
    bottomClip = nDisplayedRows - 1;
  }

  if( tlcell.GetRow() > bottomClip || brcell.GetRow() < topClip ||
      brcell.GetCol() < leftClip ){
    return 0;
  }

  int bDrawTop = 1;    // TRUE if the top of the rectangle should be drawn
  int bDrawLeft = 1;   // TRUE if the left of the rectangle should be drawn
  int bDrawBottom = 1; // TRUE if rectangle's bottom should be drawn

  /* Determine the rectangle's corners, and determine whether the 
  ** top or bottom lines should be drawn or clipped.
  */
  if( tlcell.GetRow() < topClip ){
    tlcell.SetRow(topClip);
    bDrawTop = 0;
  }
  if( tlcell.GetCol() < leftClip ){
    tlcell.SetCol(leftClip);
    bDrawLeft = 0;
  }
  SsBbox *pTopBbox = Bbox(tlcell);
  if( !pTopBbox ){
    return 0;
  }
  int x0 = pTopBbox->x0;
  int y0 = pTopBbox->y0;
  delete pTopBbox;

  if( brcell.GetRow() > bottomClip ){
    brcell.SetRow(bottomClip);
    bDrawBottom = 0;
  }
  SsBbox *pBottomBbox = Bbox(brcell);
  if( !pBottomBbox ){
    return 0;
  }
  int x1 = pBottomBbox->x1;
  int y1 = pBottomBbox->y1;
  delete pBottomBbox;

  SsLineCommand *pCmd = fastcan.CreateLineCommand(zTag, zColor);
  pCmd->SetWidth(width);
  if( bDrawLeft ){
    pCmd->SetCoords(x0, y0, x0, y1); pCmd->Execute();
  }
  pCmd->SetCoords(x1, y0, x1, y1); pCmd->Execute();
  if( bDrawTop ){
    pCmd->SetCoords(x0, y0, x1, y0); pCmd->Execute();
  }
  if( bDrawBottom ){
    pCmd->SetCoords(x0, y1, x1, y1); pCmd->Execute();
  }
  delete pCmd;
  return 1;
}

/*
** This method will make the background rectangle (which is used to simulate
** boundaries) a bit smaller.  This is a trick that will make
** the "find closest" command return a cell or text glyph rather than
** the background rectangle, when the coordinate is outside the spreadsheet.
*/
PRIVATE void ss::ShrinkBgrect(){
  int id=fastcan.FindFirst("bgrect");
  if( id==0 ) return;
  fastcan.Eval("coords %d", id);
  float x0, x1, y0, y1;
  sscanf(evalStrResult(), "%f %f %f %f", &x0, &y0, &x1, &y1);
  if( x1>0 && y1>0 ){
    x1 -= 1.0;
    y1 -= 1.0;
    fastcan.Eval("coords %d %f %f %f %f", id, x0, y0, x1, y1);
  }
}

/*
** This method will restore the background rectangle back to its original
** shape.  See ShrinkBgrect() above.
*/
PRIVATE void ss::RestoreBgrect(){
  int id=fastcan.FindFirst("bgrect");
  if( id==0 ) return;
  fastcan.Eval("coords %d", id);
  float x0, x1, y0, y1;
  sscanf(evalStrResult(), "%f %f %f %f", &x0, &y0, &x1, &y1);
  if( x1>0 && y1>0 ){
    x1 += 1.0;
    y1 += 1.0;
    fastcan.Eval("coords %d %f %f %f %f", id, x0, y0, x1, y1);
  }
}

/*
** Get the highest row that should not be displayed in the spreadsheet.
** If bTrim is nonzero, rows that do not contain data will 
** be trimmed from the bottom.
** Care will be taken to avoid leaving an unused area at the bottom.
**
** In this method, the word "highest" refers to the position in the spreadsheet,
** not the row number.  The smaller a row number, 
** the higher it appears in the spreadsheet.
*/
PRIVATE int ss::GetHighestUndisplayedRow(int bTrim) {
  int scrollableHeight = GetScrollableHeight();
  int topRow = min(max(topleft.GetRow(), 1), sConfigure.maxRows - 1);

  /*
  ** The highest undisplayed row is at least 
  ** the maximum of the first row just below the
  ** spreadsheet, and configured minimum number of rows.
  **
  ** We have to change nDisplayedRows to do these calculations.  Its value
  ** will be restored on exit.
  */
  int nSaveDisplayedRows = nDisplayedRows;
  nDisplayedRows = sConfigure.maxRows;
  int highestUndisplayedRow = 
      max(GetRowAtDistance(topRow, scrollableHeight) + 1, sConfigure.rows);

  if( bTrim ){  // trim off rows without data from the bottom
    highestUndisplayedRow = max(highestUndisplayedRow, GetLastRowWithData()+1);
    highestUndisplayedRow = min(highestUndisplayedRow, sConfigure.maxRows);
  } else {
    highestUndisplayedRow = max(highestUndisplayedRow, nSaveDisplayedRows);
  }

  /* It is still possible that there is unused space at the 
  ** bottom of the spreadsheet.  Fill in that space.
  */
  int lowestToprow = 
      GetRowWithinDistance(highestUndisplayedRow, -scrollableHeight);
  highestUndisplayedRow = 
      GetRowWithinDistance(lowestToprow, scrollableHeight) + 1;

  nDisplayedRows = nSaveDisplayedRows;
  return highestUndisplayedRow;
}

/*
** Scroll the spreadsheet so the given "cell" is visible.
*/
PRIVATE void ss::MakeCellVisible(SsCell &cell){
  int scrollableHeight = GetScrollableHeight();
  int scrollableWidth = GetScrollableWidth();
  int lastRow = GetRowWithinDistance(topleft.GetRow(), scrollableHeight);
  int lastCol = GetColWithinDistance(topleft.GetCol(), scrollableWidth);

  int row = cell.GetRow();
  if( row<1 ){
    row = 1;
  }
  if( row >= sConfigure.maxRows ){
    row = sConfigure.maxRows - 1;
  }

  int col = cell.GetCol();
  if( col<1 ){
    col = 1;
  }
  if( col >= sConfigure.cols ){
    col = sConfigure.cols - 1;
  }
  if( row >= topleft.GetRow() && row <= lastRow &&
     col >= topleft.GetCol() && col <= lastCol ) return;

  /* If the row to see is above the viewport, scroll so
  ** that the row to see is the highest visible one.
  ** If the row to see is below the viewport, scroll so that the row to
  ** see is the lowest visible one.
  ** Do a similar operation for the column.
  */
  if( row < topleft.GetRow() ){
    evalf("%s yview %d", zSs, row - 1);
  } else if( row > lastRow ){
    int nSaveDisplayedRows = nDisplayedRows;
    nDisplayedRows = sConfigure.maxRows;
    int topRow = GetRowWithinDistance(row + 1, -scrollableHeight);
    nDisplayedRows = nSaveDisplayedRows;
    evalf("%s yview %d", zSs, topRow - 1);
  }

  if( col < topleft.GetCol() ){
    evalf("%s xview %d", zSs, col - 1);
  } else if( col > lastCol ){
    int leftCol = GetColWithinDistance(col + 1, -scrollableWidth);
    evalf("%s xview %d", zSs, leftCol - 1);
  }
}

/*
** This method calculates the width of the visible canvas (the "viewport")
** and returns it.
*/
PRIVATE int ss::CalculateViewportWidth(){
  fastcan.Eval("cget -width");
  int calcWidth = evalIntResult();
  if( calcWidth > 0 ){
    return calcWidth;
  }
  int nCols = 0;
  if( sConfigure.width > 0 ){
    nCols = sConfigure.width;
  } else if( sConfigure.cols > 0 ){
    nCols = sConfigure.cols;
  }

  for( int col = 0; col < nCols; ++col ){
    calcWidth += GetColumnWidth(col);
  }
  return calcWidth;
}

/*
** This method calculates the height of the visible canvas (the "viewport")
** and returns it.
*/
PRIVATE int ss::CalculateViewportHeight(){
  fastcan.Eval("cget -height");
  int calcHeight = evalIntResult();
  if( calcHeight > 0 ){
    return calcHeight;
  }
  int nConfiguredRows = sConfigure.rows;
  if( nConfiguredRows > sConfigure.maxRows ){
    nConfiguredRows = sConfigure.maxRows;
  }
  int nRows = 0;
  if( sConfigure.height > 0 ){
    nRows = sConfigure.height;
  } else if( nConfiguredRows > 0 ){
    nRows = nConfiguredRows;
  }
  if( nRows > 0 ){
    calcHeight = GetConfiguredHeight(0) + sConfigure.rowHeight*(nRows-1);

    /* Now take into account the rows with specifically configured heights.
    */
    for( SsArrayElem *pElem=aHeight.first(); pElem; pElem=pElem->next() ){
      int row = (int)(pElem->key());
      if( row < nRows && row > 0 ){
        int rowHeight = (int)(pElem->data());
        calcHeight -= sConfigure.rowHeight - rowHeight;
      }
    }
  }
  return calcHeight;
}

/*
** Determine whether the cell's text value is too large to fit, and if so,
** reset the height of the cell's row.
*/
PRIVATE void ss::ResetHeight(SsCell &cell){
  SsString *pText = fastcan.Itemcget(cell.render(), "-text");
  if( !pText->isnull() ){
    int colWidth = GetColumnWidth(cell.GetCol());
    int newHeight = GetTextHeight(pText->render(), colWidth);
    if( newHeight > GetConfiguredHeight(cell.GetRow()) ){
      evalf("%s height %d %d", zSs, cell.GetRow(), newHeight);
    }
  }

  delete pText;
}

/*
** Calculate the height of all rows in the spreadsheet in the range
** fromRow to toRow.  The row heights will be written into the aHeight array
** (for rows other than the title row), or into the 
** currentRowHeight member variable (for the title row).
**
** The "heights" entry in the widget variable may be changed by this routine.
*/
PRIVATE void ss::CalculateRowHeights(int fromRow, int toRow) {
  if( !sConfigure.zVariable || !*sConfigure.zVariable ) return;

  int anyHeightSet = 0;
  int *pWidths = new int[sConfigure.cols];
  for( int col = 0; col<sConfigure.cols; ++col ){
    pWidths[col] = GetColumnWidth(col);
  }
  for( int row=fromRow; row <= toRow; row++ ){
    int rowTextHeight = GetRowTextHeight(row, pWidths);
    if( row==0 ){
      if( rowTextHeight>sConfigure.titleRowHeight ){
        currentTitleRowHeight = rowTextHeight;
      } else {
        currentTitleRowHeight = -1;
      }
    } else {
      if( rowTextHeight > GetConfiguredHeight(row) ){
        aHeight.insert((void *)row, (void *)rowTextHeight);
        anyHeightSet++;
      }
    }
  }
  delete [] pWidths;

  /* Reset the "heights" entry in the widget variable.
  */
  if( anyHeightSet ){
    evalf("%s height", zSs);
    char *zRes = (char *)evalStrResult();
    const char *zHeights = GetWidgetVariable("heights");
    if( !zHeights || strcmp(zHeights, zRes)!=0 ){
      SetWidgetVariable("heights", zRes);
    }
  }
}

/*
** Return the minimum height that "row" must have in order to fit the
** text contained in each cell within "row", using the current width of each
** column.
*/
PRIVATE int ss::GetRowTextHeight(int row, int *pWidths){
  int maxHeight = 0;  // maximum height of all cells visited so far

  char zCell[66];
  sprintf(zCell, "%d,", row);
  char *zCellRow = &zCell[strlen(zCell)];
  for( int col=0; col < sConfigure.cols; col++ ){
    if( col<10 ){
      zCellRow[0] = '0' + col;
      zCellRow[1] = 0;
    } else {
      zCellRow[0] = '0' + col/10;
      zCellRow[1] = '0' + col%10;
      zCellRow[2] = 0;
    }
    const char *zText = GetWidgetVariable(zCell);
    int width = *pWidths++;
    int theight = ( zText && *zText ) ?  GetTextHeight(zText, width) : 0;
    if( theight > maxHeight ){
      maxHeight = theight;
    }
  }
  return maxHeight;
}

/*
** This method returns the height that the text string zText will require
** when it appears in a cell whose width is "colWidth".  The returned value
** includes a couple of extra pixels so that text has a pixel of pad on
** either side.
*/
PRIVATE int ss::GetTextHeight(const char *zText, int colWidth){
  int pad = cellPad*2;
  if( !zText || !*zText || colWidth<=pad ){
    return 0;
  }

  int twidth, theight;
  Tk_TextLayout textLayout = Tk_ComputeTextLayout(font,
      zText, -1, colWidth-pad, TK_JUSTIFY_LEFT, 0, &twidth, &theight);
  Tk_FreeTextLayout(textLayout);
  return theight + 2;
}

/*
** Return the highest numbered row that contains displayable data.
** (Note that the row containing the active cell is considered to have
** displayable data.)
** The row will be cached in the member named lastRowWithData.
** If the spreadsheet cells haven't changed, then the cached value is returned.
*/
PRIVATE int ss::GetLastRowWithData() {
  if( !bDirtyCell ){
    return lastRowWithData;
  }
  int highestRow = activeCell.GetRow();
  if( highestRow<0 ) highestRow = 0;

  if( sConfigure.zVariable && *sConfigure.zVariable ){
    evalf("array names %s", sConfigure.zVariable);
    Tcl_Obj *pList = Tcl_GetObjResult(interp);
    if( pList ){
      int objc;
      Tcl_Obj **objv;
      if( Tcl_ListObjGetElements(interp, pList, &objc, &objv)==TCL_OK ){

        /* Note that only non-title elements count.
        */
        for( int i=0; i<objc; i++ ){
          Tcl_Obj *pElt = objv[i];
          char *zString = Tcl_GetString(pElt);
          int row, col;
          if( sscanf(zString, "%d,%d", &row, &col)==2 && row>highestRow 
              && row<sConfigure.maxRows && col!=0 ){
            highestRow = row;
          }
        }
      }
    }
  }
  lastRowWithData = highestRow;
  return highestRow;
}

/*
** Return the distance in pixels between "col1" and "col2".
*/
PRIVATE int ss::GetDistanceBetweenCols(int col1, int col2){
  assert( col1 >= 1 && col1 <= sConfigure.cols );
  assert( col2 >= 1 && col2 <= sConfigure.cols );
  int c1 = min(col1, col2);
  int c2 = max(col1, col2);
  int pix = 0;
  while( c1 < c2 ){
    pix += GetColumnWidth(c1++);
  }
  if( col2 < col1 ){
    pix *= -1;
  }
  return pix;
}

/*
** Return the column that is "distance" pixels away from column "col".
** The return value is constrained so that the return value is 
** between 1, and the number of columns minus one.
**
** Rounding is performed so that the distance between the start of "col"
** and the furthest point of the returned column is at most "distance".
*/
PRIVATE int ss::GetColWithinDistance(int col, int distance){
  int newCol = col;   // the column to find
  int incr = 1;       // the direction from the original column to the new one

  if( distance < 0 ){
    assert(col > 0 && col <= sConfigure.cols);
    distance *= -1;
    incr = -1;
    newCol--;
  } else {
    assert(col > 0 && col < sConfigure.cols);
  }
  int distanceSoFar = 0;    // distance from initial column, so far
  while( newCol > 0 && newCol < sConfigure.cols ){
    int newDistance = distanceSoFar + GetColumnWidth(newCol);
    if( newDistance > distance ) break;
    newCol += incr;
    distanceSoFar = newDistance;
  }
  if( incr < 0 ){
    newCol++;
  } else if( newCol != col ){
    newCol--;
  }
  return newCol;
}

/*
** This proc is like GetColWithinDistance(), but this one does not
** round.  In other words, it returns the column displayed at "distance" pixels
** from the start of "col".
*/
PRIVATE int ss::GetColAtDistance(int col, int distance){
  int distanceSoFar;  // Number of pixels traversed so far
  int incr;           // The direction from the original column to the new one

  if( distance < 0 ){
    assert(col > 0 && col <= sConfigure.cols);
    distance *= -1;
    incr = -1;
    distanceSoFar = 0;
  } else {
    assert(col > 0 && col < sConfigure.cols);
    incr = 1;
    distanceSoFar = GetColumnWidth(col);
  }
  int newCol = col;
  while( newCol > 0 && newCol < sConfigure.cols && distanceSoFar < distance){
    newCol += incr;
    distanceSoFar += GetColumnWidth(newCol);
  }
  if( newCol==0 ){
    newCol++;
  }
  if( newCol==sConfigure.cols ){
    newCol--;
  }
  return newCol;
}

/*
** Return the distance in pixels between "row1" and "row2".
*/
PRIVATE int ss::GetDistanceBetweenRows(int row1, int row2){
  assert( row1 >= 1 && row1 <= sConfigure.maxRows );
  assert( row2 >= 1 && row2 <= sConfigure.maxRows );
  int r1 = min(row1, row2);
  int r2 = max(row1, row2);
  int pix = 0;
  while( r1 < r2 ){
    pix += GetConfiguredHeight(r1++);
  }
  if( row2 < row1 ){
    pix *= -1;
  }
  return pix;
}

/*
** Return the row that is "distance" pixels away from "row".
** The return value is constrained so that the return value is 
** between 1, and the number of displayed rows minus one.
**
** Rounding is performed so that the distance between the start of "row"
** and the furthest point of the returned row is at most "distance".
*/
PRIVATE int ss::GetRowWithinDistance(int row, int distance){

  int newRow = row;   // the row to find
  int incr = 1;       // the direction from the original row to the new one
  if( distance < 0 ){
    assert(row > 0 && row <= nDisplayedRows);
    distance *= -1;
    incr = -1;
    newRow--;
  } else {
    assert(row > 0 && row < nDisplayedRows);
  }
  int distanceSoFar = 0;    // distance from initial row, so far
  int *pWidths = new int[sConfigure.cols];
  for( int col = 0; col<sConfigure.cols; ++col ){
    pWidths[col] = GetColumnWidth(col);
  }

  /* The actual row height may not have yet been calculated yet, because
  ** it isn't calculated until the row is drawn.  So take the maximum of the
  ** configured height (which is a default value for an undrawn row)
  ** and the text height (which will be its actual drawn height).
  */
  while( newRow > 0 && newRow < nDisplayedRows ){
    int newDistance = distanceSoFar + 
      max(GetConfiguredHeight(newRow), GetRowTextHeight(newRow, pWidths));
    if( newDistance > distance ) break;
    newRow += incr;
    distanceSoFar = newDistance;
  }
  delete [] pWidths;
  if( incr < 0 ){
    if( newRow != row-1 ){
      newRow++;
    }
  } else {
    if( newRow != row ){
      newRow--;
    }
  }
  return newRow;
}

/*
** This method is like ss::GetRowWithinDistance(), but this one does not
** round.  In other words, it returns the row displayed at "distance" pixels
** from the start of "row".
*/
PRIVATE int ss::GetRowAtDistance(int row, int distance){
  int incr = 1;       // the direction from the original row to the new one
  int distanceSoFar;  // Number of pixels traversed so far

  int *pWidths = new int[sConfigure.cols];
  for( int col = 0; col<sConfigure.cols; ++col ){
    pWidths[col] = GetColumnWidth(col);
  }

  /* The actual row height may not have yet been calculated yet, because
  ** it isn't calculated until the row is drawn.  So take the maximum of the
  ** configured height (which is a default value for an undrawn row)
  ** and the text height (which will be its actual drawn height).
  */
  if( distance < 0){
    assert(row > 0 && row <= nDisplayedRows);
    distance *= -1;
    incr = -1;
    distanceSoFar = 0;
  } else {
    assert(row > 0 && row < nDisplayedRows);
    incr = 1;
    distanceSoFar = 
        max(GetConfiguredHeight(row), GetRowTextHeight(row, pWidths));
  }
  int newRow = row;
  while( newRow > 0 && newRow < nDisplayedRows && distanceSoFar < distance ){
    newRow += incr;
    distanceSoFar += 
        max(GetConfiguredHeight(newRow), GetRowTextHeight(newRow, pWidths));
  }
  delete [] pWidths;
  if( newRow==0 ){
    newRow++;
  }
  if( newRow==nDisplayedRows ){
    newRow--;
  }
  return newRow;
}

/*
** Recolor the foreground and background color of the given "cell".
*/
PRIVATE void ss::RecolorCell(SsCell &cell){
  if( displayProperties & DISPLAY_PENDING ){  // redisplay already pending
    return;
  }
  if( cell.isnull() ) return;
  char *zCell = cell.render();
  SsString cellTag("C"); cellTag.append(zCell);
  fastcan.Itemconfigure(cellTag.render(), "-fill", GetCellBg(cell));
  fastcan.Itemconfigure(zCell, "-fill", GetCellFg(cell));
}

/*
** Recolor just the background of the given "cell"
*/
PRIVATE void ss::RecolorCellBg(SsCell &cell){
  if( displayProperties & DISPLAY_PENDING ){  // redisplay already pending
    return;
  }
  if( cell.isnull() ) return;
  SsString cellTag("C"); cellTag.append(cell.render());
  fastcan.Itemconfigure(cellTag.render(), "-fill", GetCellBg(cell));
}

/*
** Recolor just the foreground of the given "cell"
*/
PRIVATE void ss::RecolorCellFg(SsCell &cell){
  if( displayProperties & DISPLAY_PENDING ){ // redisplay already pending
    return;
  }
  if( cell.isnull() ) return;
  fastcan.Itemconfigure(cell.render(), "-fill", GetCellFg(cell));
}

/*
** Return the background color for the given cell
*/
PRIVATE char *ss::GetCellBg(SsCell &cell){
  if( (cell.GetRow()==0 || cell.GetCol()==0) && 
      sConfigure.zTitleBg && *sConfigure.zTitleBg ){
    return sConfigure.zTitleBg;
  }
  if( sConfigure.zActiveBg && *sConfigure.zActiveBg && cell.equal(activeCell) ){
    return sConfigure.zActiveBg;
  }

  char *zCell = cell.render();
  if( sConfigure.zSelectBg && *sConfigure.zSelectBg && 
      fastcan.CellIsSelected(zCell) ){
    return sConfigure.zSelectBg;
  }
  char *zCellBg = (char *) aBg[(void *)zCell];
  if( zCellBg && *zCellBg ){
    return zCellBg;
  }
  return sConfigure.zBg;
}

/*
** Return the foreground color for the given cell
*/
PRIVATE char *ss::GetCellFg(SsCell &cell){
  if( (cell.GetRow()==0 || cell.GetCol()==0) && 
      sConfigure.zTitleFg && *sConfigure.zTitleFg ){
    return sConfigure.zTitleFg;
  }
  if( sConfigure.zActiveFg && *sConfigure.zActiveFg && cell.equal(activeCell) ){
    return sConfigure.zActiveFg;
  }

  char *zCell = cell.render();
  if( sConfigure.zSelectFg && *sConfigure.zSelectFg &&
      fastcan.CellIsSelected(zCell) ){
    return sConfigure.zSelectFg;
  }
  char *zCellFg = (char *)aFg[(void *)zCell];
  if( zCellFg && *zCellFg ){
    return zCellFg;
  }
  return sConfigure.zFg;
}

/*
** Recolor the foreground and background color of all cells in the 
** spreadsheet.  This is often faster than doing a set of
** RecolorCell()'s, for a group of cells.
** Even so, this proc is rather slow, so it should be used sparingly.
*/
PRIVATE void ss::RecolorAllCells(){
  RecolorAllBackgrounds();
  RecolorAllForegrounds();
}

/* 
** Recolor the background of every non-title cell in the spreadsheet.
*/
PRIVATE void ss::RecolorAllBackgrounds(){
  if( displayProperties & DISPLAY_PENDING ){ // redisplay already pending
    return;
  }

  /* First, the default background.
  */
  fastcan.Eval("itemconfigure cell -fill %s", sConfigure.zBg);

  SsItemconfigCommand *pBgCmd = 
      fastcan.CreateItemconfigCommand("-fill", "");

  /* Next, the per-cell background.
  */
  for( SsArrayElem *pElem = aBg.first(); pElem; pElem = pElem->next() ){
    char *zCellBg = (char *)(pElem->data());
    if( *zCellBg ){
      SsString cellTag("C"); cellTag.append((char *)(pElem->key()));
      int id = fastcan.TagToFirstId(cellTag.render());
      if( id!=0 ){
        pBgCmd->SetId(id);
        pBgCmd->SetValue(zCellBg);
        pBgCmd->Execute();
      }
    }
  }
  delete pBgCmd;
  
  /* Finally, the selected and active backgrounds.
  */
  ColorSelectedBg();
  if( sConfigure.zActiveBg && *sConfigure.zActiveBg ){
    fastcan.Itemconfigure("activecell", "-fill", sConfigure.zActiveBg);
  }
}

/*
** Configure the foreground color of every non-title spreadsheet cell.
*/
PRIVATE void ss::RecolorAllForegrounds(){
  if( displayProperties & DISPLAY_PENDING ){ // redisplay already pending
    return;
  }

  /* First, the default foreground.
  */
  if( sConfigure.zFg && *sConfigure.zFg ){
    fastcan.Eval("itemconfigure text -fill %s", sConfigure.zFg);
  }

  /* Next, the per-cell foregrounds.
  */
  SsItemconfigCommand *pFgCmd = 
      fastcan.CreateItemconfigCommand("-fill", "");
  for( SsArrayElem *pElem = aFg.first(); pElem; pElem = pElem->next() ){
    char *zCellFg = (char *)(pElem->data());
    if( *zCellFg ){
      pFgCmd->SetId(fastcan.TagToFirstId((char *)(pElem->key())));
      pFgCmd->SetValue(zCellFg);
      pFgCmd->Execute();
    }
  }
  delete pFgCmd;

  /* Finally, the selected and active foregrounds.
  */
  ColorSelectedFg();
  if( sConfigure.zActiveFg && *sConfigure.zActiveFg ){
    fastcan.Itemconfigure("active", "-fill", sConfigure.zActiveFg);
  }
}

/* Color the given "cell" with the selection color, if it isn't active
** and if the selection colors are configured.
*/
PRIVATE void ss::SelectCell(SsCell &cell){
  if( cell.GetRow()<=0 || cell.GetCol()<=0 ) return;

  int bIsActive = cell.equal(activeCell);
  if( sConfigure.zSelectBg && *sConfigure.zSelectBg &&
      (!sConfigure.zActiveBg || !*sConfigure.zActiveBg || !bIsActive) ){
    SsString cellTag("C"); cellTag.append(cell.render());
    fastcan.Itemconfigure(cellTag.render(), "-fill", sConfigure.zSelectBg);
  }

  if( sConfigure.zSelectFg && *sConfigure.zSelectFg &&
      (!sConfigure.zActiveFg || !*sConfigure.zActiveFg || !bIsActive) ){
    fastcan.Itemconfigure(cell.render(), "-fill", sConfigure.zSelectFg);
  }
}

/* Color all of the cells and cell data in the spreadsheet with the 
** background and foreground selection colors, respectively,
** except for the active cell, whose colors override the selection color.
**
** The active cell will always be recolored by this routine (unless a
** redisplay is pending).
*/
PRIVATE void ss::ColorAllSelected(){
  if( displayProperties & DISPLAY_PENDING ){ // redisplay already pending
    return;
  }
  char *zBg = (sConfigure.zSelectBg && *sConfigure.zSelectBg) ?
    sConfigure.zSelectBg : 0;
  char *zFg = (sConfigure.zSelectFg && *sConfigure.zSelectFg) ?
    sConfigure.zSelectFg : 0;
  if( zBg || zFg ){
    void *pHandle;
    for( char *zCell = fastcan.FirstSelected(&pHandle); zCell;
        zCell = fastcan.NextSelected(&pHandle) ){
      if( zBg ){
        SsString cellTag("C"); cellTag.append(zCell);
        fastcan.Itemconfigure(cellTag.render(), "-fill", sConfigure.zSelectBg);
      }
      if( zFg ){
        fastcan.Itemconfigure(zCell, "-fill", sConfigure.zSelectFg);
      }
    }
  }
  if( sConfigure.zActiveBg && *sConfigure.zActiveBg ){
    fastcan.Itemconfigure("activecell", "-fill", sConfigure.zActiveBg);
  }
  if( sConfigure.zActiveFg && *sConfigure.zActiveFg ){
    fastcan.Itemconfigure("active", "-fill", sConfigure.zActiveFg);
  }
}

/*
** Color all selected cells with the selected background color.
*/
PRIVATE void ss::ColorSelectedBg(){
  if( sConfigure.zSelectBg && *sConfigure.zSelectBg ){
    void *pHandle;
    for( char *zCell = fastcan.FirstSelected(&pHandle); zCell;
        zCell = fastcan.NextSelected(&pHandle) ){
      SsString cellTag("C"); cellTag.append(zCell);
      fastcan.Itemconfigure(cellTag.render(), "-fill", sConfigure.zSelectBg);
    }
  }
}

/*
** Color all selected text with the selected foreground color.
*/
PRIVATE void ss::ColorSelectedFg(){
  if( sConfigure.zSelectFg && *sConfigure.zSelectFg ){
    void *pHandle;
    for( char *zCell = fastcan.FirstSelected(&pHandle); zCell;
        zCell = fastcan.NextSelected(&pHandle) ){
      fastcan.Itemconfigure(zCell, "-fill", sConfigure.zSelectFg);
    }
  }
}

/*
** Deselect the given cell by changing its foreground and background colors
** to the appropriate deselected color.
*/
PRIVATE void ss::DeselectCell(SsCell &cell){
  if( cell.GetRow()<=0 || cell.GetCol()<=0 ) return;

  /* First, reset the cell's background color
  */
  char *zCell = cell.render();
  char *zCellBg = (char *) aBg[(void *)zCell];
  int bIsActive = cell.equal(activeCell);
  char *zColor;

  if( bIsActive && sConfigure.zActiveBg && *sConfigure.zActiveBg ){
    zColor = sConfigure.zActiveBg;
  } else if( zCellBg && *zCellBg ){
    zColor = zCellBg;
  } else {
    zColor = sConfigure.zBg;
  }
  SsString cellTag("C"); cellTag.append(zCell);
  fastcan.Itemconfigure(cellTag.render(), "-fill", zColor);

  /* Next reset the cell's foreground color
  */
  char *zCellFg = (char *) aFg[(void *)zCell];
  if( bIsActive&& sConfigure.zActiveFg && *sConfigure.zActiveFg ){
    zColor = sConfigure.zActiveFg;
  } else if( zCellFg && *zCellFg ){
    zColor = zCellFg;
  } else {
    zColor = sConfigure.zFg;
  }
  fastcan.Itemconfigure(zCell, "-fill", zColor);
}

/*
** This method will de-select every cell in the spreadsheet.
*/
PRIVATE void ss::DeselectAllCells(){
  fastcan.DeselectAll();
  if( sConfigure.zSelectBg && *sConfigure.zSelectBg ){
    RecolorAllBackgrounds();
  }
  if( sConfigure.zSelectFg && *sConfigure.zSelectFg ){
    RecolorAllForegrounds();
  }
}

/*
** Return the current height of the scrollable portion of the viewport
*/
PRIVATE int ss::GetScrollableHeight(){
  fastcan.Eval("cget -height");
  int fullHeight = evalIntResult();
  return fullHeight - GetConfiguredHeight(0);
}

/*
** Return the current width of the scrollable portion of the viewport
*/
PRIVATE int ss::GetScrollableWidth(){
  fastcan.Eval("cget -width");
  int fullWidth = evalIntResult();
  return fullWidth - GetColumnWidth(0);
}

/*
** Given the "row", return its Y coordinate within the canvas.
** Return -1 if the coordinate could not be calculated.
*/
PRIVATE int ss::GetRowY(int row){
  assert(row > 0 && row <= nDisplayedRows);
  SsCell cell(row-1, 0);
  SsString cellTag("C"); cellTag.append(cell.render());
  int id = fastcan.TagToFirstId(cellTag.render());
  if( id==0 ){
    return -1;
  }
  fastcan.Eval("bbox %d", id);
  SsBbox bbox(evalStrResult());
  return bbox.y1;
}

PRIVATE int ss::min(int a, int b){
  return (a < b ? a : b);
}
PRIVATE int ss::max(int a, int b){
  return (a > b ? a : b);
}
