/*
** This module implements hashed tags on canvas items.
** Tags provided natively by the canvas widget are not hashed, meaning that
** looking up a canvas id by tag (using e.g. "find withtag") is linear.
** If the spreadsheet is very large, this lookup is very slow.
** So we provide our own implementation of tags that uses hash tables to
** provide constant time lookup.
**
** There are a few native canvas tags that we do utilize.  These tags use
** canvas bindings to implement row and column resizing.
** See for example the tags "hresize", "vresize", "notresize", and "title".
** We do not want to re-implement canvas bindings in the ss widget.
**
** We also use native canvas tags if the expected number of canvas id's with
** the tag is large relative to the total number of id's.  For example,
** the "cell" and "text" tags are native canvas tags, because each represents
** about 1/2 of the total number of id's in the spreadsheet.  There is
** little or no benefit to using a hashed tag, when an iterator would have
** to visit nearly every canvas id anyway.
*/
static char const rcsid[] = "@(#) $Header: /home/softools/cvs/softools/code/common/cked/src/ss_fastcan.cpp,v 1.20 2004/04/05 19:32:40 mark Exp $";

#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <tcl.h>
#include <tk.h>
#include "ss_fastcan.h"

#if INTERFACE

/*
** object:  SsCanvas
**
** A single object of the SsCanvas class represents a "fast canvas" that
** implements linear tag lookup, among other things, for good performance.
*/
class SsCanvas {
private:
  Tcl_Interp *interp;
  SsSetElem *pPrevTagElem; // previous tag element looked up
  SsSetElem *pPrevIdElem;  // previous canvas id element looked up

  SsArrayOfPointer aIdToTags; // Array mapping canvas id's to list of tags
  SsArrayOfPointer aIdToCell; // Array mapping canvas id's to cell
  SsArrayOfString aSelection; // Array containing all currently selected cells.
  SsArrayOfString aTagToId;   // Array mapping a canvas tag to a single 
                              // canvas id

  SsArrayOfStaticString aTagToIdSet; // Array mapping canvas tag to a set of 
                                     // canvas id's.
                                     // This is used to allow very fast lookup 
                                     // from a canvas tag to a list of canvas 
                                     // ids.  It is used in the cases where a 
                                     // tag may be defined on many canvas id's.
public:
  char *zName;  // the name of the actual canvas widget

  SsCanvas(){ interp = 0; zName = 0; pPrevTagElem = 0; pPrevIdElem = 0;}
  SsCanvas(Tcl_Interp *pI, char *zN){ 
    init(pI, zN); 
    pPrevTagElem = 0;
    pPrevIdElem = 0;
  }

  /*
  ** Initialize the canvas by creating its underlying Tk representation
  ** and by adding the "Ss" class bindings to it.  Change the name of
  ** the widget proc by appending the suffix ":can" to it.
  */
  void SsCanvas::init(Tcl_Interp *pI, const char *zN){
    interp = pI;
    evalf("canvas %s -width 0 -height 0", zN);
    evalf("bindtags %s [concat [bindtags %s] Ss]", zN, zN);
    delete [] zName;
    zName = new char[strlen(zN) + 7];
    sprintf(zName, "::%s:can", zN);
    evalf("rename ::%s ::%s", zN, zName);

    /* Initialize the map of canvas tags to sets of canvas id's.
    ** 
    ** There are three types of tags that are used in the spreadsheet:
    ** - Native canvas tags (e.g. "hresize", "vresize", "cell", "text");
    ** - "Fast canvas" tags that map a single tag to a set of canvas id's; and
    ** - "Fast canvas" tags that map a single tag to a single canvas id.
    **
    ** Native canvas tags are used for one of two reasons:
    ** Either the tag has a tag binding (and we did not wish to implement
    ** tag bindings in the "fast canvas" implementation), or the tag is used
    ** for a significant portion of the total number of canvas id's
    ** ("cell" and "text" each are used by roughly half of all canvas id's).
    ** 
    ** "Fast canvas" tags are used in the other cases.  The are needed because
    ** the native canvas implementation of tags uses a linear lookup when
    ** finding the canvas id's associated with a tag.  We do not want to
    ** traverse a list of 10000 canvas id's just to find a single canvas id.
    ** So we create hash tables that map tags to canvas id's.
    ** See ss_fastcan.cpp for more details of the "fast canvas" implementation.
    **
    ** The following code initializes the "fast canvas" hash tables that map a
    ** single tag to multiple canvas id's.  The one-to-one mapping of
    ** tags to canvas id's is in the associative array aTagToId.
    */
    aTagToIdSet.insert((void *)"mx", new SsSetOfPointer);
    aTagToIdSet.insert((void *)"my", new SsSetOfPointer);
    aTagToIdSet.insert((void *)"resize", new SsSetOfPointer);
    aTagToIdSet.insert((void *)"title", new SsSetOfPointer);
    aTagToIdSet.insert((void *)"titlecell", new SsSetOfPointer);
  }

  ~SsCanvas(){

    /* Order of deletion is important here.
    ** First delete the elements in the associative arrays, then delete
    ** the arrays themselves.
    */
    DeleteAll();
    delete (SsSetOfPointer *)aTagToIdSet.find((void*)"mx");
    delete (SsSetOfPointer *)aTagToIdSet.find((void*)"my");
    delete (SsSetOfPointer *)aTagToIdSet.find((void*)"resize");
    delete (SsSetOfPointer *)aTagToIdSet.find((void*)"title");
    delete (SsSetOfPointer *)aTagToIdSet.find((void*)"titlecell");
    delete [] zName;
  }
}

/*
** object: SsItemconfigCommand
**
** A single object of the SsItemconfigCommand class will be 
** used to create and execute a canvas "itemconfigure" command, very quickly.
*/
class SsItemconfigCommand : public SsTclList {
public:
  SsItemconfigCommand(Tcl_Interp *pInterp, char *zWidgetCmd, 
      const char *zOption, const char *zValue) : SsTclList(pInterp) {
    append(zWidgetCmd);
    append("itemconfigure");
    append(0);       // element #2 will be filled in later
    append(zOption);
    append(zValue);  // element #4 might be filled in later
  }

  void SetId(int id){
    lset(2, id);
  }
  void SetValue(char *zValue){
    lset(4, zValue);
  }
};

/*
** object: SsCellCommand
**
** A single object of the SsCellCommand class defines a Tcl command that will
** draw a cell (rectangle) on a spreadsheet.
** This class is defined so that a single Tcl command can be reused 
** multiple times in a loop, rather than forcing a command to be 
** re-parsed many times.
*/
class SsCellCommand : public SsTclList {
public:
  SsCellCommand(Tcl_Interp *pInterp, const char *zWidgetCmd, const char *zTags, 
      const char *zFill) : SsTclList(pInterp) {
    append(zWidgetCmd);
    append("create");
    append("rectangle");
    append(0);         // elements #3 thru #6 will be filled in later
    append(0);
    append(0);
    append(0);
    append("-fill");
    append(zFill);     // element #8 might be filled in later
    append("-tags");
    append(zTags);
    append("-outline");
    append("");
  }

  void SetCoords(int x0, int y0, int x1, int y1){
    lset(3, x0);
    lset(4, y0);
    lset(5, x1);
    lset(6, y1);
  }
  void SetXCoords(int x0, int x1){
    lset(3, x0);
    lset(5, x1);
  }
  void SetYCoords(int y0, int y1){
    lset(4, y0);
    lset(6, y1);
  }

  void SetFill(char *zFill){ lset(8, zFill); }
};

/*
** object: SsTextCommand
**
** A single object of the SsTextCommand class defines a Tcl command that will
** draw a text item on a spreadsheet.
** This class is defined so that a single Tcl command can be reused 
** multiple times in a loop, rather than forcing a command to be 
** re-parsed many times.
**
** [SRS4010] [SRS4020]
** The text width is defined so that the text word-wraps to the
** size of the cell that contains it.
*/
class SsTextCommand : public SsTclList {
public:
  SsTextCommand(Tcl_Interp *pInterp, const char *zWidgetCmd, const char *zFont, 
      const char *zTags, const char *zAnchor, const char *zFill) : 
      SsTclList(pInterp) {
    append(zWidgetCmd);
    append("create");
    append("text");
    append(0);     // element #3:  will be filled in later
    append(0);     // element #4:  will be filled in later
    append("-text");
    append("");    // element #6:  will be filled in later
    append("-width");
    append(0);     // element #8:  will be filled in later
    append("-fill");
    append(zFill); // element #10: might be filled in later
    append("-tags");
    append(zTags);
    append("-anchor");
    append(zAnchor); 
    append("-font");
    append(zFont);
  }

  void SetCoords(int x, int y){
    lset(3, x);
    lset(4, y);
  }
  void SetXCoord(int x){ lset(3, x); }
  void SetYCoord(int y){ lset(4, y); }
  void SetText(const char *zText){ lset(6, zText); }
  void SetWidth(int width) { lset(8, width); }
  void SetFill(char *zFill){ lset(10, zFill); }
};

/*
** object: SsLineCommand
**
** A single object of the SsLineCommand class defines a Tcl command that
** will draw a line item on a spreadsheet.
** This class is defined so that a single Tcl command can be reused 
** multiple times in a loop, rather than forcing a command to be 
** re-parsed many times.
*/
class SsLineCommand : public SsTclList {
public:
  SsLineCommand(Tcl_Interp *pInterp, const char *zWidgetCmd, const char *zTags, 
      const char *zFill) : SsTclList(pInterp) {
    append(zWidgetCmd);
    append("create");
    append("line");
    append(0);                 // elements #3 thru #6 will be filled in later
    append(0);
    append(0);
    append(0);
    append("-fill");
    append(zFill);
    append("-width");
    append(0);                 // element #10 may be filled in later
    append("-stipple");
    append("");                // element #12 may be filled in later
    append("-tags");
    append(zTags);
  }

  void SetWidth(int width){
    lset(10, width);
  }
  void SetCoords(int x0, int y0, int x1, int y1){
    lset(3, x0);
    lset(4, y0);
    lset(5, x1);
    lset(6, y1);
  }
  void SetXCoords(int x0, int x1){
    lset(3, x0);
    lset(5, x1);
  }
  void SetYCoords(int y0, int y1){
    lset(4, y0);
    lset(6, y1);
  }
  void SetStipple(const char *zStipple){
    lset(12, zStipple);
  }
};
#endif

/*
** Add the tag zTagToAdd to the target tag referenced by zTargetTag.
*/
PUBLIC void SsCanvas::AddTag(const char *zTagToAdd, const char *zTargetTag){
  for( int id=TagToFirstId(zTargetTag); id!=0; id=TagToNextId(zTargetTag) ){
    AddTag(zTagToAdd, id);
  }
}

/*
** Overload the AddTag function to account for the case where the
** target tag is just a single canvas id.
*/
PUBLIC void SsCanvas::AddTag(const char *zTagToAdd, int id){
  MapTagToId(zTagToAdd, id);
  MapIdToTag(id, zTagToAdd);
}

/*
** Add a tag that is known to be "one to one" relative to the given id.
*/
PUBLIC void SsCanvas::AddOneToOneTag(const char *zTagToAdd, int id){
  MapOneToOneTagToId(zTagToAdd, id);
  MapIdToTag(id, zTagToAdd);
}

/*
** Add the tag in zTagToAdd to every canvas Id in the given "list".
*/
PUBLIC void SsCanvas::AddTagToList(const char *zTagToAdd, SsTclList &list){
  int id;
  for( int ret = list.first_int(&id); ret!=0; ret = list.next_int(&id) ){
    MapTagToId(zTagToAdd, id);
    MapIdToTag(id, zTagToAdd);
  }
}

/*
** Return a bounding box for the canvas items that possess all of the
** canvas tags referenced by pTagSet.
** The returned bounding box must be freed by the caller.
** If there are no canvas items referenced by the tag set, 
** this method returns 0.
*/
PUBLIC SsBbox *SsCanvas::Bbox(const SsSetOfString *pTagSet){
  SsTclList idList(interp);
  for( SsSetElem *pElem=pTagSet->first(); pElem; pElem=pElem->next() ){
    char *zTag = (char *)(pElem->key());  // Use key() for sets
    for( int id=TagToFirstId(zTag); id!=0; id=TagToNextId(zTag) ){
      idList.append(id);
    }
  }
  if( idList.isnull()) {
    return 0;
  }
  if( Eval("bbox %s", idList.render())==TCL_OK ) {
    return new SsBbox(evalStrResult());
  } else {
    return 0;
  }
}

/*
** For each canvas item possessing zTag, delete the characters in
** the range given by zFrom and zTo, inclusively.
** But if the zTo is 0, make zTo the same as zFrom.
*/
PUBLIC void SsCanvas::Dchars(const char *zTag, char *zFrom, char *zTo){
  if( !zFrom ){
    return;
  }
  if( !zTo ){
    zTo = zFrom;
  }
  int iValue;
  if( strcmp(zFrom, "end")!=0 && sscanf(zFrom, "%d", &iValue)!=1 ){
    return;
  }
  if( strcmp(zTo, "end")!=0 && sscanf(zTo, "%d", &iValue)!=1 ){
    return;
  }
  for( int id=TagToFirstId(zTag); id!=0; id=TagToNextId(zTag) ){
    Eval("dchars %d %s %s", id, zFrom, zTo);
  }
}

/*
** Same as the above, but allow "from" and "to" to be integers.
*/
PUBLIC void SsCanvas::Dchars(const char *zTag, int from, int to){
  for( int id=TagToFirstId(zTag); id!=0; id=TagToNextId(zTag) ){
    Eval("dchars %d %d %d", id, from, to);
  }
}

/*
** Delete each item given by the space-separated
** list of canvas id's referenced by zIdList.
*/
PUBLIC void SsCanvas::DeleteIds(const char *zIdList){
  SsString list = zIdList;
  for( char *zId = list.first(); zId; zId = list.next() ){
    int id = atoi(zId);
    for( char *zTag = IdToFirstTag(id); zTag; zTag = IdToNextTag() ){
      UnmapTagToId(zTag, id);
    }
    UnmapIdToAllTags(id);
    UnmapIdToCell(id);
  }
  Eval("delete %s", zIdList);
}

/*
** This method deletes all tags and ids from the spreadsheet
*/
PUBLIC void SsCanvas::DeleteAll(){
  ClearTagToIds();
  ClearIdToTags();
  ClearIdToCell();
}

/*
** For each canvas id that possesses zTag, delete the tag given by 
** zTagToDelete.
*/
PUBLIC void SsCanvas::Dtag(const char *zTag, const char *zTagToDelete){
  for( int id=TagToFirstId(zTag); id!=0; id=TagToNextId(zTag) ){
    UnmapIdToTag(id, zTagToDelete);
    UnmapTagToId(zTagToDelete, id);
  }
}

/*
** Delete the tag pointed to by zTagToDelete from all of the canvas id's.
*/
PUBLIC void SsCanvas::DtagAll(const char *zTagToDelete){
  for( int id=TagToFirstId(zTagToDelete); id!=0; id=TagToNextId(zTagToDelete) ){
    UnmapIdToTag(id, zTagToDelete);
  }
  UnmapTagToAllIds(zTagToDelete);
}

/*
** Return the first canvas id that possesses zTag.
** The list of canvas id's is in no particular order.
**
** The caller should make sure to traverse only one list of canvas id's 
** at a time.
*/
PUBLIC int SsCanvas::FindFirst(const char *zTag){
  return TagToFirstId(zTag);
}

/*
** Return the next canvas id that possesses zTag.
** The list of canvas id's is in no particular order.
**
** The caller should make sure to traverse only one list of canvas id's 
** at a time.
*/
PUBLIC int SsCanvas::FindNext(const char *zTag){
  return TagToNextId(zTag);
}

/*
** Return the number of canvas id's that possess zTag.
*/
PUBLIC int SsCanvas::GetIdCount(const char *zTag){
  return TagToIdCount(zTag);
}

/*
** Set the keyboard focus for the canvas widget to the first canvas item
** that possesses the tag zTag.  If zTag is 0, then the focus is reset
** so that no canvas id has the focus.
*/
PUBLIC void SsCanvas::Focus(char *zTag){
  if( zTag ){
    int id = TagToFirstId(zTag);
    if( id!=0 ){
      Eval("focus %d", id);
    }
  } else {
    Eval("focus {}");
  }
}

/*
** Return nonzero if the first canvas id possessing zTagId also possesses
** the tag given by zTag.
*/
PUBLIC int SsCanvas::HasTag(const char *zTagId, const char *zTag){
  int id = TagToFirstId(zTagId);
  if( id==0 ){
    return 0;
  }
  return HasTag(id, zTag);
}

/* The same as the above, but this takes an id as an argument instead of
** a tag.
*/
PUBLIC int SsCanvas::HasTag(int id, const char *zTag){
  for( char *zOldTag = IdToFirstTag(id); zOldTag; zOldTag = IdToNextTag() ){
    if( *zTag==*zOldTag && strcmp(zTag, zOldTag)==0 ){
      return 1;
    }
  }
  return 0;
}

/*
** Set the position of the insertion cursor for the canvas id(s) that
** possess the tag referenced by zTag,
** to just before the character whose position is referenced by zIndex.
*/
PUBLIC void SsCanvas::Icursor(const char *zTag, const char *zIndex){
  for( int id=TagToFirstId(zTag); id!=0; id=TagToNextId(zTag) ){
    Eval("icursor %d %s", id, zIndex);
  }
}
/*
** Overload the above method to allow for integral indices.
*/
PUBLIC void SsCanvas::Icursor(const char *zTag, const int index){
  for( int id=TagToFirstId(zTag); id!=0; id=TagToNextId(zTag) ){
    Eval("icursor %d %d", id, index);
  }
}

/*
** Return a decimal string giving the numerical index within the first
** canvas id possessing tag zTag corresponding to the index specifier zIndex.
** Return -1 if there is no such index.
*/
PUBLIC int SsCanvas::Index(const char *zTag, const char *zIndex){
  return Index(TagToFirstId(zTag), zIndex);
}

/*
** Return a decimal string giving the numerical index for the given
** canvas "id" corresponding to the index specifier zIndex.
** Return -1 if there is no such index.
*/
PUBLIC int SsCanvas::Index(const int id, const char *zIndex){
  if( id==0 ){
    return -1;
  } else {
    if( Eval("index %d %s", id, zIndex)==TCL_OK ){
      return evalIntResult();
    } else {
      return -1;
    }
  }
}

/*
** Insert zString at the insertion point of the active text element.
** This proc will create a new text glyph for the active text element
** if one does not already exist in the active cell.
**
** The text glyph is created even if zString is 0 or "".  This is necessary
** in order to make the cursor blink when the user enters the (empty) cell.
*/
PUBLIC void SsCanvas::InsertActive(const char *zString, ss *pSs){
  int id = TagToFirstId("active");
  if( id==0 ){
    int cellId = TagToFirstId("activecell");
    if( cellId==0 ) return;
    id = pSs->CreateTextGlyph(cellId, zString);
    if( id!=0 ){
      SsCell *pCell = IdToCell(cellId);
      char *zCell = pCell->render();
      Focus(zCell);
      Icursor(zCell, "end");
    }
  } else {
    Eval("insert %d insert \"%q\"", id, zString);
  }
}

/*
** Return a pointer to a SsString object that contains the value for the 
** option given by zOption, for the first canvas id that possesses zTag.
** Return a string referencing "" if there is no such canvas id or option.
** The return is malloc'ed; the caller must free it.
*/
PUBLIC SsString *SsCanvas::Itemcget(const char *zTag, const char *zOption){
  SsString *pNew = new SsString();
  int id = TagToFirstId(zTag);
  if( id!=0 && Eval("itemcget %d %s", id, zOption)==TCL_OK ){
    pNew->set(evalStrResult());
  }
  return pNew;
}

/*
** Configure every canvas id possessing zTag so that its option named
** zOption is given the value named zValue.
** If zValue is 0, then the option is given the value "".
*/
PUBLIC void SsCanvas::Itemconfigure(const char *zTag, const char *zOption, 
    const char *zValue){
  SsItemconfigCommand *pCmd = new SsItemconfigCommand(interp, zName,
      zOption, zValue ? zValue : "");

  for( int id=TagToFirstId(zTag); id!=0; id=TagToNextId(zTag) ){
    pCmd->SetId(id);
    pCmd->Execute();
  }
  delete pCmd;
}

/*
** This method is similar to Itemconfigure(), but it just sets the
** value of the text object whose name is zCell.
** It is a special case because it may be necessary to create the text glyph
** on the fly.
**
** The text glyph is created even if zValue is 0 or "".  This is necessary
** in order to make the cursor blink when the user enters the (empty) cell.
*/
PUBLIC void SsCanvas::SetText(const char *zCell, const char *zValue, ss *pSs){
  int id = TagToFirstId(zCell);
  if( id!=0 ){
    Itemconfigure(zCell, "-text", zValue);
  } else if( pSs ){
    char zCellTag[64];
    sprintf(zCellTag, "C%s", zCell);
    int cellId = TagToFirstId(zCellTag);
    if( cellId==0 ) return;
    pSs->CreateTextGlyph(cellId, zValue ? zValue : "");
  }
}

/*
** Move each of the canvas ids possessing the tag zTag in the canvas coordinate
** space by adding "xAmount" to the x-coordinate of each point associated
** with the item, and "yAmount" to the y-coordinate of each point associated
** with the item.
*/
PUBLIC void SsCanvas::Move(const char *zTag, const int xAmount, 
    const int yAmount){
  for( int id=TagToFirstId(zTag); id!=0; id=TagToNextId(zTag) ){
    Eval("move %d %d %d", id, xAmount, yAmount);
  }
}

/*
** A single instance of this class will be used to create and execute
** a canvas "raise" command, very quickly.
*/
class SsRaiseCommand : public SsTclList {
private:
  int objc;
  Tcl_Obj **objv;
public:
  SsRaiseCommand(Tcl_Interp *pInterp, char *zWidgetCmd) : SsTclList(pInterp) {
    append(zWidgetCmd);
    append("raise");
    append(0); // element #2 will be filled in later
    Tcl_ListObjGetElements(interp, get_listrep(), &objc, &objv);
  }

  void SetId(int id){
    Tcl_SetIntObj(objv[2], id);
  }
  int Execute(){ return Tcl_EvalObjv(interp, objc, objv, 0); }
};

/*
** Move each of the canvas ids possessing the tag zTag to the highest
** position in the canvas display list.
*/
PUBLIC void SsCanvas::Raise(const char *zTag){
  SsRaiseCommand *pRaiseCommand = new SsRaiseCommand(interp, zName);
  for( int id=TagToFirstId(zTag); id!=0; id=TagToNextId(zTag) ){
    pRaiseCommand->SetId(id);
    pRaiseCommand->Execute();
  }
  delete pRaiseCommand;
}

/*
** Locate the end of the text selection in the first canvas id 
** possessing tag zTag nearest to the character indicated by zIndex, 
** and adjust that end of the selection to be at zIndex.
*/
PUBLIC void SsCanvas::SelectAdjust(const char *zTag, const char *zIndex){
  int id = TagToFirstId(zTag);
  if( id!=0 ){
    Eval("select adjust %d %s", id, zIndex);
  }
}

/*
** Set the text selection anchor point for the canvas widget to be just
** before the character given by zIndex in first the canvas id that
** possesses zTag
*/
PUBLIC void SsCanvas::SelectFrom(const char *zTag, const char *zIndex){
  int id = TagToFirstId(zTag);
  if( id!=0 ){
    Eval("select from %d %s", id, zIndex);
  }
}
PUBLIC void SsCanvas::SelectFrom(const char *zTag, int idx){
  int id = TagToFirstId(zTag);
  if( id!=0 ){
    Eval("select from %d %d", id, idx);
  }
}

/*
** Set the text selection to consist of those characters between
** the selection anchor and zIndex, for the canvas id possessing zTag.
*/
PUBLIC void SsCanvas::SelectTo(const char *zTag, const char *zIndex){
  int id = TagToFirstId(zTag);
  if( id!=0 ){
    Eval("select to %d %s", id, zIndex);
  }
}
PUBLIC void SsCanvas::SelectTo(const char *zTag, int idx){
  int id = TagToFirstId(zTag);
  if( id!=0 ){
    Eval("select to %d %d", id, idx);
  }
}

/*
** The following methods are "factory" methods.  They will create objects
** used to evaluate Tcl commands efficiently.
*/
PUBLIC SsCellCommand *SsCanvas::CreateCellCommand(
  const char *zTags, const char *zFill
){
  return new SsCellCommand(interp, zName, zTags, zFill);
}

PUBLIC SsTextCommand *SsCanvas::CreateTextCommand(
  const char *zFont, const char *zTags, const char *zAnchor, const char *zFill
){
  return new SsTextCommand(interp, zName, zFont, zTags, zAnchor, zFill);
}

PUBLIC SsLineCommand *SsCanvas::CreateLineCommand(
  const char *zTags, const char *zFill
){
  return new SsLineCommand(interp, zName, zTags, zFill);
}

PUBLIC SsItemconfigCommand *SsCanvas::CreateItemconfigCommand(
  const char *zOption, const char *zValue
){
  return new SsItemconfigCommand(interp, zName, zOption, zValue);
}

/*
** The code below implements the selection interface.
** The selection interface is not managed using fast canvas tags, because the
** selection must persist even if the underlying canvas glyphs are removed
** because of a redisplay operation.
*/

/* 
** This method is used to add a single cell to the selection.
*/
PUBLIC void SsCanvas::SelectCell(const char *zCell){
  aSelection.insert((void *)zCell, (void *)1);
}

/*
** This method is used to add all cells to the selection.
*/
PUBLIC void SsCanvas::SelectAll(int nrows, int ncols){
  for( int row=1; row<nrows; ++row ){
    for( int col=1; col<ncols; ++col ){
      SsCell cell(row, col);
      SelectCell(cell.render());
    }
  }
}

/*
** This method deletes a cell from the selection.
*/
PUBLIC void SsCanvas::DeselectCell(const char *zCell){
  aSelection.remove(zCell);
}

/*
** This method deletes every cell from the selection.
*/
PUBLIC void SsCanvas::DeselectAll(){
  aSelection.reset();
}

/*
** This method returns TRUE if the given cell is in the selection.
*/
PUBLIC int SsCanvas::CellIsSelected(const char *zCell){
  return (int) aSelection.find((void *)zCell);
}

/*
** This method returns a pointer to a sorted SsTclList containing all
** selected cells.  The list is malloc'ed; the caller must free it.
*/
PUBLIC SsTclList *SsCanvas::GetSortedSelection(){
  SsTclList *pSelection = GetSelection();
  evalf("lsort -dictionary \"%s\"", pSelection->render());
  delete pSelection;
  return new SsTclList(interp, Tcl_GetObjResult(interp));
}

/*
** This method returns a pointer to an unsorted SsTclList containing all
** selected cells.  The list is malloc'ed; the caller must free it.
*/
PUBLIC SsTclList *SsCanvas::GetSelection(){
  SsTclList *pSelection = new SsTclList(interp);
  for( SsArrayElem *p=aSelection.first(); p; p=p->next() ){
    char *zCell = (char *)(p->key());
    pSelection->append(zCell);
  }
  return pSelection;
}

/*
** This method returns the first selected cell on the list of selected cells.
** The list is unsorted, so the meaning of "first" is arbitrary.
** In addition, the "handle" of the list is returned in pHandle.  This can
** be used if the NextSelected() method is to be called later.
**
** The method returns 0 if there are no members on the list.
*/
PUBLIC char *SsCanvas::FirstSelected(void **ppHandle){
  SsArrayElem *p = aSelection.first();
  *ppHandle = (void *)p;
  if( p ){
    return (char *)(p->key());
  } else {
    return 0;
  }
}

/*
** This method returns the next selected cell on the list of selected cells.
** See also FirstSelected().
*/
PUBLIC char *SsCanvas::NextSelected(void **ppHandle){
  SsArrayElem *p = (SsArrayElem *)(*ppHandle);
  p = p->next();
  *ppHandle = (void *)p;
  if( p ){
    return (char *)(p->key());
  } else {
    return 0;
  }
}

/*
** Evaluate the 'varargs' Tcl command for the fast canvas widget.
*/
PUBLIC int SsCanvas::Eval(const char *zFormat, ...){
  va_list ap;
  va_start(ap, zFormat);
  char *zCmd = vmprintf(zFormat,ap);
  va_end(ap);

  char *zNewCmd = new char[strlen(zName) + strlen(zCmd) + 2];
  sprintf(zNewCmd, "%s %s", zName, zCmd);
  Tcl_Free(zCmd);
  int result = Tcl_Eval(interp, zNewCmd);
  delete [] zNewCmd;
  return result;
}

/*
** This method returns the first canvas id in the list of canvas id's that
** possess the given tag.  It returns 0 if there is no canvas id for the tag.
** (This implies that 0 is not a valid canvas id, which is the case.)
**
** SIDE EFFECT: pPrevIdElem is modified by this routine.
** The caller must be careful to traverse only one list of 
** canvas id elements at a time.
*/
PUBLIC int SsCanvas::TagToFirstId(const char *zTag){

  /* First, look up the tag in the one-to-one list; if it's not there,
  ** look it up in the one-to-many list
  */
  int id = (int) aTagToId[(void *)zTag];
  if( id==0 ){
    pPrevIdElem = 0;
    SsSetOfPointer *pSetOfIds = (SsSetOfPointer *)aTagToIdSet[(void *)zTag];
    if( pSetOfIds ){
      SsSetElem *pElem = pSetOfIds->first();
      if( pElem ){
        id = (int) pElem->key(); // Use key() for sets, not data()
      }
      pPrevIdElem = pElem;
    }
  }
  return id;
}

/*
** This method returns the next canvas id in the list of canvas id's that
** possess the given tag.  It returns 0 if there are no more id's for the tag.
** (This implies that 0 is not a valid canvas id, which is the case.)
**
** SIDE EFFECT: pPrevIdElem is modified by this routine.
** The caller must be careful to traverse only one list of 
** canvas id elements at a time.
*/
PUBLIC int SsCanvas::TagToNextId(const char *zTag){

  /* First, look up the tag in the one-to-one list; if it's not there,
  ** look it up in the one-to-many list
  */
  int id = (int) aTagToId[(void *)zTag];
  if( id!=0 ){  // there can be only one id for zTag on the one-to-one table
    return 0;
  }
  id = 0;
  if( pPrevIdElem ){
    pPrevIdElem = pPrevIdElem->next();
    if( pPrevIdElem ){
      id = (int) pPrevIdElem->key();  // Use key() for sets, not data()
    }
  }
  return id;
}

/*
** Given a canvas id, return the first tag that it possesses.
** Return 0 if it doesn't possess any.
**
** SIDE EFFECT: ppPrevTagElem is modified by this routine.
** The caller must be careful to traverse only one list of 
** tags at a time.
*/
PUBLIC char *SsCanvas::IdToFirstTag(const int id){
  SsSetOfString *pSetOfTags = (SsSetOfString *)aIdToTags[(void *)id];
  if( !pSetOfTags ){
    return 0;
  }
  pPrevTagElem = pSetOfTags->first();
  return (char *)(pPrevTagElem->key()); // Use key() for sets, not data()
}

/*
** Given a canvas id, return the next tag in the list of tags that it
** possesses.  Return 0 if there are no more tags in the list.
**
** SIDE EFFECT: ppPrevTagElem is modified by this routine.
** The caller must be careful to traverse only one list of 
** tags at a time.
*/
PUBLIC char *SsCanvas::IdToNextTag(){
  if( !pPrevTagElem ){
    return 0;
  }
  pPrevTagElem = pPrevTagElem->next();
  if( pPrevTagElem ){
    return (char *)(pPrevTagElem->key());  // Use key() for sets, not data()
  } else {
    return 0;
  }
}

/*
** This method maps the canvas id to its spreadsheet cell.
** This is a many-to-one mapping; that is, there can be several canvas id's
** that map to the same cell.
**
** A private copy of the cell is made.
*/
PUBLIC void SsCanvas::MapIdToCell(const int id, const SsCell &cell){
  SsCell *pNewCell = new SsCell(cell);
  SsCell *pOldCell = (SsCell *)aIdToCell.insert((void *)id, (void *)pNewCell);
  delete pOldCell;
}

/*
** This is just like the last method, except it takes the row and column of
** the cell as input.
*/
PUBLIC void SsCanvas::MapIdToCell(const int id, int row, int col){
  SsCell *pNewCell = new SsCell(row, col);
  SsCell *pOldCell = (SsCell *)aIdToCell.insert((void *)id, (void *)pNewCell);
  delete pOldCell;
}

/*
** Remove the mapping from the id to its spreadsheet cell.
*/
PUBLIC void SsCanvas::UnmapIdToCell(const int id){
  SsCell *pCell = (SsCell *) aIdToCell.remove((void *)id);
  if( pCell ){
    delete pCell;
  }
}

/*
** Return a pointer to the cell that the canvas id is mapped to.
*/
PUBLIC SsCell *SsCanvas::IdToCell(const int id){
  return (SsCell *)aIdToCell[(void *)id];
}

/*
** All of the methods below are private to the SsCanvas class.
*/

/*
** This method maps the tag pointed to by zTag, to the given canvas id.
*/
PRIVATE void SsCanvas::MapTagToId(const char *zTag, const int id){

  /* First we determine if this tag has a one-to-many lookup table.
  ** If so, we use that one for the mapping; otherwise, we put the
  ** map in the one-to-one lookup table.
  */
  SsSetOfPointer *pSetOfIds = (SsSetOfPointer *)aTagToIdSet[(void *)zTag];
  if( pSetOfIds ){
    pSetOfIds->insert((void *)id);
  } else {
    aTagToId.insert((void *)zTag, (void *)id);
  }
}

/*
** This method maps the tag pointed to by zTag to the given canvas id.
** The tag is known to be one-to-one with respect to the given id.
*/
PRIVATE void SsCanvas::MapOneToOneTagToId(const char *zTag, const int id){
  aTagToId.insert((void *)zTag, (void *)id);
}

/*
** This method unmaps the mapping from zTag to the given canvas id
*/
PRIVATE void SsCanvas::UnmapTagToId(const char *zTag, const int id){

  /* Determine whether the mapping is on the one-to-many or the one-to-one
  ** lookup table.  Delete it from the appropriate lookup table.
  */
  SsSetOfPointer *pSetOfIds = (SsSetOfPointer *)aTagToIdSet[(void *)zTag];
  if( pSetOfIds ){
    pSetOfIds->remove((void *)id);
  } else {
    aTagToId.remove((void *)zTag);
  }
}

/*
** This method unmaps the mapping between zTag and all canvas id's it is
** mapped to
*/
PRIVATE void SsCanvas::UnmapTagToAllIds(const char *zTag){
  SsSetOfPointer *pSetOfIds = (SsSetOfPointer *)aTagToIdSet[(void *)zTag];
  if( pSetOfIds ){
    pSetOfIds->reset();
  } else {
    aTagToId.remove((void *)zTag);
  }
}

/*
** This method clears both the one-to-one and one-to-many mappings
** from tags to ids
*/
PRIVATE void SsCanvas::ClearTagToIds(){
  ClearOneToOneTagToId();
  ClearOneToManyTagToId();
}

/*
** This method clears the one-to-one mapping from tags to ids
*/
PRIVATE void SsCanvas::ClearOneToOneTagToId(){
  aTagToId.reset();
}

/*
** This method clears the one-to-many mapping from tags to ids.
*/
PRIVATE void SsCanvas::ClearOneToManyTagToId(){
  for( SsArrayElem *pElem = aTagToIdSet.first(); pElem; 
      pElem = pElem->next() ) {
    SsSetOfPointer *pIdSet = (SsSetOfPointer *)(pElem->data());
    pIdSet->reset();
  }

  /* Note that the associative array aTagToIdSet itself is not reset.
  ** Its keys (the names of designated tags) are constants.
  */
}

/*
** This method returns a single canvas id that possesses the given tag.
** It returns 0 if there is no such canvas id.
** (This implies that 0 is not a valid canvas id, which is the case.)
*/
PRIVATE int SsCanvas::TagToOneId(const char *zTag){

  /* First the one-to-one mapping table is used for the lookup.
  ** If the tag isn't in that table, then the one-to-many mapping table
  ** is used instead.
  */
  int id = (int) aTagToId[(void *)zTag];
  if( id==0 ){
    SsSetOfPointer *pSetOfIds = (SsSetOfPointer *)aTagToIdSet[(void *)zTag];
    if( pSetOfIds ){
      SsSetElem *pElem = pSetOfIds->first();
      if( pElem ){
        id = (int) pElem->key();  // Use key() for sets, not data()
      }
    }
  }
  return id;
}

/*
** Return the number of canvas id's that possess zTag
*/
PRIVATE int SsCanvas::TagToIdCount(const char *zTag){
  int id = (int) aTagToId[(void *)zTag];
  if( id!=0 ){  // there must be only one id for zTag on the one-to-one table
    return 1;
  }
  SsSetOfPointer *pSetOfIds = (SsSetOfPointer *)aTagToIdSet[(void *)zTag];
  if( pSetOfIds ){
    return pSetOfIds->count();
  } else {
    return 0;
  }
}

/*
** This method maps the canvas id to the given tag.
** This is a one-to-many mapping.
*/
PRIVATE void SsCanvas::MapIdToTag(const int id, const char *zTag){
  SsSetOfString *pSetOfTags = (SsSetOfString *)aIdToTags[(void *)id];
  if( !pSetOfTags ){
    pSetOfTags = new SsSetOfString;
    aIdToTags.insert((void *)id, pSetOfTags);
  }
  pSetOfTags->insert((void *)zTag);
}

/*
** Remove the mapping from the canvas id to its list of tags
*/
PRIVATE void SsCanvas::UnmapIdToAllTags(const int id){
  SsSetOfString *pSetOfTags = (SsSetOfString *)aIdToTags[(void *)id];
  if( pSetOfTags ){
    aIdToTags.remove((void *)id);
    delete pSetOfTags;
  }
}

/*
** Remove just the given tag pointed to by zTag from the one-to-many
** mapping from canvas id to tags.
*/
PRIVATE void SsCanvas::UnmapIdToTag(const int id, const char *zTag){
  SsSetOfString *pSetOfTags = (SsSetOfString *)aIdToTags[(void *)id];
  if( pSetOfTags ){
    pSetOfTags->remove((void *)zTag);
  }
}

/*
** This method deletes the id to tag mapping and frees all memory
*/
PRIVATE void SsCanvas::ClearIdToTags() {
  for( SsArrayElem *pElem = aIdToTags.first(); pElem; pElem = pElem->next() ) {
    delete (SsSetOfString *)(pElem->data());
  }
  aIdToTags.reset();
}

/*
** This proc is used for debugging
*/
PRIVATE void SsCanvas::DumpIdToCell(){
  for( SsArrayElem *pElem = aIdToCell.first(); pElem; pElem = pElem->next() ) {
    SsCell *pCell = (SsCell *)(pElem->data());
    fprintf(stderr, "id %d -> cell %s\n", (int)pElem->key(), pCell->render());
  }
}

/*
** This method deletes the entire id to spreadsheet cell mapping
*/
PRIVATE void SsCanvas::ClearIdToCell(){
  for( SsArrayElem *pElem = aIdToCell.first(); pElem; pElem = pElem->next() ) {
    delete (SsCell *)(pElem->data());
  }
  aIdToCell.reset();
}

/*
** Evaluate the Tcl script given by the variable number of arguments
*/
PRIVATE int SsCanvas::evalf(const char *zFormat, ...){
  va_list ap;
  int result;
  va_start(ap,zFormat);
  result = ::evalf(interp, zFormat, ap);
  va_end(ap);
  return result;
}

/*
** Return the string result of the most recent Tcl command.
*/
PRIVATE const char *SsCanvas::evalStrResult(){
  return ::evalStrResult(interp);
}

/*
** Return the integer result of the most recent Tcl command.
*/
PRIVATE int SsCanvas::evalIntResult(){
  return ::evalIntResult(interp);
}
